import numpy as np
from . import Data
import scipy

class Tilt(Data):
    """
    Class that represents the tilt geodetic datatype
    
    Attributes:
        dx: inclination in the x-axis (radians)
        dy: inclination in the y-axis (radians)
        errx: uncertainty in the inclination in the x-axis (radians)
        erry: uncertainty in the inclination in the y-axis (radians)
        azx: azimuth for the x-axis, the angle is positive clockwise starting from the north
    """
    def __init__(self):
        self.names=None
        self.dx=None
        self.dy=None
        self.errx=None
        self.erry=None
        self.azx=None
        super().__init__()
    
    def add_azx(self,azx):
        """
        Adds the azimuth angles for the x-axes
        
        Parameters:
            azx (array): azimuth angles clockwise from north
        """
        self.assert_size(azx,'azx')
        self.azx=azx
        
    def add_dx(self,dx):
        """
        Adds the inclination in the x-axis
        
        Parameters:
            dx (array): inclination in the x-axis
        """
        self.add_comp(dx,'dx')
        self.dx=dx
    
    def add_errx(self,errx):
        """
        Adds the uncertainties in the inclination for the x-axis
        
        Parameters:
            errx (array): uncertainties for the inclination in the x-axis
        """
        self.add_errcomp(errx,'dx')
        self.errx=errx
    
    def add_dy(self,dy):
        """
        Adds the inclination in the y-axis
        
        Parameters:
            dy (array): inclination in the y-axis
        """
        self.add_comp(dy,'dy')
        self.dy=dy
        
    def add_erry(self,erry):
        """
        Adds the uncertainties in the inclination for the y-axis
        
        Parameters:
            erry (array): uncertainties for the inclination in the y-axis
        """
        self.add_errcomp(erry,'dy')
        self.erry=erry
        
    def get_index(self,name):
        """
        Defines the order for the components in the data array
        
        Returns:
            index (int): component's index in the data array
        """
        if name=='dx':
            return 0
        elif name=='dy':
            return 1
        
    def ref_possible(self):
        """
        Overrides the function to specify that the tilt dataset cannot take references
        """
        return False
    
    def add_data(self,dx,dy):
        """
        Adds the two components at the same time
        
        Parameters:
            dx (array): inclination in the x-axis
            dy (array): inclination in the y-axis
        """
        self.add_ux(dx)
        self.add_uy(dy)
        super().add_data(self,np.concatenate((dx,dy)))
    
    def add_err(self,errx,erry):
        """
        Adds the uncertainties for the two components at the same time
        
        Parameters:
            errx (array): uncertainty for the inclination in the x-axis
            erry (array): uncertainty for the inclination in the y-axis
        """
        self.add_errx(errx)
        self.add_erry(erry)
        super().add_err(self,np.concatenate((errx,erry)))
    
    def from_model3d(self,func,unravel=True):
        """
        Uses the function from the forward model to compute the inclination in the tiltmeters
        
        Parameters:
            func: tilt displacements forward model function
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
            
        Returns:
            model (array): array containing the components
        """
        if 'tilt' in func.__name__:
            dx,dy=func(self.xs,self.ys)
        else:
            uzx= lambda xpos: func(xpos,self.ys)[2]
            uzy= lambda ypos: func(self.xs,ypos)[2]

            dx=-scipy.misc.derivative(uzx,self.xs,dx=1e-6)
            dy=-scipy.misc.derivative(uzy,self.ys,dx=1e-6)
        
        model=()
        if isinstance(self.dx,(list,np.ndarray)):
            model=(*model,dx)
        if isinstance(self.dy,(list,np.ndarray)):
            model=(*model,dy)
        
        return model


from . import Data
import numpy as np

class Edm(Data):
    """
    Class that represents the Electronic distance measurements datatype
    
    Attributes:
        delta: deformation in the lines
        lonsori: longitudes for the lines origins
        xorigins: x-coordinates for the line origins
        lonsends: longitudes for the lines target
        xends: x-coordinates for the line ends
        latsori: latitudes for the lines origins
        yorigins: y-coordinates for the line origins
        latsends: latitudes for the lines target
        yends: y-coordinates for the line ends
        zorigins: z-coordinates for the line origins
        zends: z-coordinates for the line ends
    """
    def __init__(self):
        self.delta=None
        self.lonsori=None
        self.xorigins=None
        self.lonsend=None
        self.xends=None
        self.latsori=None
        self.yorigins=None
        self.latsend=None
        self.yends=None
        self.zorigins=None
        self.zends=None
        super().__init__()
        
    def add_ts(self,ts):
        """
        Add time steps for the datapoints
        
        Parameters:
            ts: time steps (has to be the same size as the lines)
        """
        self.assert_size(np.concatenate((ts,ts)),'ts')
        self.ts=ts
        
    def add_deltas(self,delta):
        """
        Add changes in the EDM lines
        
        Parameters:
            delta: changes in the lines (m)
        """
        delta=delta.ravel()
        self.assert_origins_ends(delta)
        self.data=delta
        self.err=delta*0+1
        self.comps=[delta]
        self.delta=delta
    
    def add_xorigins(self,origins):
        """
        Add x-coordinates for the line origins
        
        Parameters:
            origins: x-coordinates for the line origins
        """
        self.assert_origins_ends(origins)
        self.xorigins=origins
        if not self.xends is None:
            self.add_xs(np.concatenate((self.xorigins,self.xends)))
        
    def add_xends(self,ends):
        """
        Add x-coordinates for the line ends
        
        Parameters:
            ends: x-coordinates for the line ends
        """
        self.assert_origins_ends(ends)
        self.xends=ends
        if not self.xorigins is None:
            self.add_xs(np.concatenate((self.xorigins,self.xends)))
        
    def add_yorigins(self,origins):
        """
        Add y-coordinates for the line origins
        
        Parameters:
            origins: y-coordinates for the line origins
        """
        self.assert_origins_ends(origins)
        self.yorigins=origins
        if not self.yends is None:
            self.add_ys(np.concatenate((self.yorigins,self.yends)))
        
    def add_yends(self,ends):
        """
        Add y-coordinates for the line ends
        
        Parameters:
            ends: y-coordinates for the line ends
        """
        self.assert_origins_ends(ends)
        self.yends=ends
        if not self.yorigins is None:
            self.add_ys(np.concatenate((self.yorigins,self.yends)))
        
    def add_lls_lines(self,lons,lats,origin=True,ori=None):
        """
        Assign spatial positions using longitudes and latitudes
        
        Parameters:
            lons (array): longitudes for the stations or datapoints
            lats (array): latitudes for the stations or datapoints
            origin (boolean): if True the lons and lats refer to the line origins, 
            if False they refer to the line ends
            ori (array): lon/lat origin for the projection of longitudes and latitudes,
            if None the mean position is the origin
        """
        if origin==True:
            self.lonsori=lons
            self.latsori=lats
        else:
            self.lonsend=lons
            self.latsend=lats
            
        if (not self.lonsori is None) and (not self.lonsend is None):
            lons=np.concatenate((self.lonsori,self.lonsend))
            lats=np.concatenate((self.latsori,self.latsend))
            self.add_lls(lons,lats,ori=ori)
            self.xorigins=self.xs[0:int(len(lons)/2)]
            self.xends=self.xs[int(len(lons)/2):len(lons)]
            self.yorigins=self.ys[0:int(len(lons)/2)]
            self.yends=self.ys[int(len(lons)/2):len(lons)]
    
    def add_zorigins(self,origins):
        """
        Add z-coordinates for the line origins
        
        Parameters:
            origins: z-coordinates for the line origins
        """
        self.assert_origins_ends(origins)
        self.zorigins=origins
        if not self.zends is None:
            self.add_zs(np.concatenate((self.zorigins,self.zends)))
        
    def add_zends(self,ends):
        """
        Add z-coordinates for the line ends
        
        Parameters:
            ends: z-coordinates for the line ends
        """
        self.assert_origins_ends(ends)
        self.zends=ends
        if not self.zorigins is None:
            self.add_zs(np.concatenate((self.zorigins,self.zends)))
        
    def get_num_lines(self):
        """
        Gives the number of lines that compose the EDM dataset
        
        Returns:
            num (int): number of lines
        """
        params=[self.xorigins,self.yorigins,self.zorigins,self.xends,self.yends,self.zends,self.ts,self.data]
        for i,l in enumerate(params):
            if isinstance(l,(list,np.ndarray)):
                if i in [len(params)-2,len(params)-1]:
                    return len(l)/self.get_size_per_point()
                else:
                    return len(l)
        return None
        
    def assert_origins_ends(self,arr):
        """
        Checks if an input array (origins or ends) has the right size
        """
        assert isinstance(arr,(list,np.ndarray)),'It should be an array'
        assert len(arr)==self.get_num_lines() or self.get_num_lines()==None,'The size does not correspond to the postions size'
        
    def from_model3d(self,func,unravel=True):
        """
        Uses the function from the forward model to compute the changes in the lines
        
        Parameters:
            func: 3d displacements forward model function
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
            
        Returns:
            deltas (array): changes (contraction or dilation) of lines (m)
        """
        ux1,uy1,uz1=func(self.xorigins,self.yorigins)
        ux2,uy2,uz2=func(self.xends,self.yends)
        
        old_dist=np.sqrt((self.xends-self.xorigins)**2+(self.yends-self.yorigins)**2+(self.zends-self.zorigins)**2)
        new_dist=np.sqrt((self.xends-self.xorigins+(ux2-ux1))**2+(self.yends-self.yorigins+(uy2-uy1))**2+(self.zends-self.zorigins+(uz2-uz1))**2)
        
        deltas=new_dist-old_dist
        return (deltas,)


import numpy as np
import os
from vmod import util


class Data:
    """
    Class to implement interface to geodetic data types used in source inversions

    Parameters:
        names: names for the datapoints (stations)
        xs: positions of datapoints in the x-axis (m)
        ys: positions of datapoints in the y-axis (m)
        lons: longitudes of datapoints (m)
        lats: latitudes of datapoints (m)
        zs: elevations of datapoints (m)
        ts: times for datapoints (s)
        comps: names for the components in the dataset (e.g., for gps, comps=['ux','uy','uz'])
        data: deformation dataset, they array is ordered according to comps (m)
        err: uncertainties in the dataset (m)
        refs: names for the datapoints that are the references
        utmz: utm zone for the projection
        
    """
    def __init__(self):
        self.names=None
        self.xs=None
        self.ys=None
        self.lons=None
        self.lats=None
        self.zs=None
        self.ts=None
        self.comps=None
        self.data=None
        self.err=None
        self.refs=None
        self.utmz=None
        #self.data_size_per_point=0
    
    def get_size_per_point(self):
        """
        Gives number of components per datapoint (time-space)
        
        Returns:
            numcomps: number of components
        """
        if self.comps==None:
            return 0
        else:
            return len(self.comps)
        
    def add_names(self,names):
        """
        Assign names to the datapoints
        
        Parameters:
            names: names for the stations or datapoints
        """
        self.assert_size(names,'names')
        self.names=names
    
    def add_lls(self,lons,lats,ori=None):
        """
        Assign spatial positions using longitudes and latitudes
        
        Parameters:
            lons: longitudes for the stations or datapoints
            lats: latitudes for the stations or datapoints
            ori: origin for the projection of longitudes and latitudes if None the mean position is the origin
        """
        xs,ys,z1s,z2s=util.ll2utm(lons,lats)
        if ori is None:
            self.utmz=[np.mean(xs),np.mean(ys),z1s,z2s]
            self.add_xs(xs-np.mean(xs))
            self.add_ys(ys-np.mean(ys))
        else:
            orix,oriy,z1sor,z2sor=util.ll2utm([ori[0]],[ori[1]],z1=z1s,z2=z2s)
            self.utmz=[orix[0],oriy[0],z1sor,z2sor]
            self.add_xs(xs-orix[0])
            self.add_ys(ys-oriy[0])
        self.lons=lons
        self.lats=lats
        
    def add_ys(self,ys):
        """
        Assign y-axis positions for the stations or datapoints
        
        Parameters:
            ys: y-coordinates for stations or datapoints
        """
        self.assert_size(ys,'ys')
        self.ys=ys
    
    def add_xs(self,xs):
        """
        Assign x-axis positions for the stations or datapoints
        
        Parameters:
            xs: x-coordinates for stations or datapoints
        """
        self.assert_size(xs,'xs')
        self.xs=xs
    
    def add_zs(self,zs):
        """
        Assign z-axis positions for the stations or datapoints
        
        Parameters:
            zs: z-coordinates for stations or datapoints
        """
        self.assert_size(zs,'zs')
        self.zs=zs
    
    def add_ts(self,ts):
        """
        Assign times for the stations or datapoints
        
        Parameters:
            ts: times for stations or datapoints
        """
        self.assert_size(ts,'ts')
        self.ts=ts
        
    def add_comp(self,comp,name):
        """
        Add component to the dataset
        
        Parameters:
            comp: data points for the component
            name: name for the component
        """
        self.assert_size(comp,name)
        i=self.get_index(name)
        if self.comps==None:
            new_data=comp
            new_err=comp*0+1
            self.comps=[name]
        else:
            replace=False
            k=self.get_order(name)
            if name in self.comps:
                print('This component was already included; it will be replaced')
                replace=True
            if replace:
                new_data=np.concatenate((self.data[0:k*len(comp)],comp,self.data[(k+1)*len(comp)::]))
                new_err=np.concatenate((self.err[0:k*len(comp)],comp*0+1.0,self.err[(k+1)*len(comp)::]))
            else:
                new_data=np.concatenate((self.data[0:k*len(comp)],comp,self.data[k*len(comp)::]))
                new_err=np.concatenate((self.err[0:k*len(comp)],comp*0+1.0,self.err[k*len(comp)::]))
                self.comps=self.comps[0:k]+[name]+self.comps[k::]
        self.data=new_data
        self.err=new_err
        
    def get_order(self,name):
        """
        Gives the position of the component in the dataset
        
        Parameters:
            name: name for the component
        """
        i=self.get_index(name)
        for k,compname in enumerate(self.comps):
            j=self.get_index(compname)
            if j>=i:
                break
            elif k==len(self.comps)-1:
                k+=1
        return k
        
    def add_errcomp(self,err,name):
        """
        Adds the uncertainties for certain component
        
        Parameters:
            err: uncertainties for the component
            name: component's name to add the uncertainties
        """
        assert name in self.comps,'The component '+name+' has not been included'
        self.assert_size(err,name)
        k=self.get_order(name)
        self.err[k*len(err):(k+1)*len(err)]=err
    
    def add_locs(self, xs, ys, ts=None, zs=None):
        """
        Adds the spatial-temporal positions for the datapoints
        
        Parameters:
            xs: x-coordinates for the datapoints
            ys: y-coordinates for the datapoints
            ts: times for the datapoints
            zs: z-coordinates for the datapoints
        """
        self.add_xs(xs)
        self.add_ys(ys)
        if isinstance(zs,(list,np.ndarray)):
            self.add_zs(zs)
        if isinstance(ts,(list,np.ndarray)):
            self.add_ts(ts)
    
    def get_index(self,name):
        """
        Gives the position for the component
        
        Parameters:
            name: component's name
        """
        return 0
    
    def add_data(self,data):
        """
        Add the deformation values to the dataset
        
        Parameters:
            data: deformation data
        """
        self.assert_size(data,'data')
        self.data=data
        
    def assert_size(self,arr,name):
        """
        Asserts the size of the array. The correspondance should be 1-1 with deformation data and uncertainties
        the correspondance of spatial datapoints with deformation data is 1-number of components
        
        Parameters:
            data: deformation data
        """
        assert isinstance(arr,(list,np.ndarray)),name+' must be an array'
        if (name=='data' or name=='errors') and not self.get_size_per_point()==0:
            assert len(arr)/self.get_size_per_point()==self.get_num_points() or self.get_num_points()==None,"The number of "+name+" points does not have the proper size"
        else:
            assert len(arr)==self.get_num_points() or self.get_num_points()==None,"The number of "+name+" points does not have the proper size"
        
    def add_err(self,err):
        """
        Adds the uncertainties for all the dataset
        
        Parameters:
            err: uncertainties for the dataset
        """
        self.assert_size(err,'errors')
        self.err=err
    
    def add(self, xs, ys, data, ts=None, zs=None):
        """
        Adds spatial-temporal locations and deformation data
        
        Parameters:
            xs: x-coordinates for the datapoints
            ys: y-coordinates for the datapoints
            data: deformation data
            zs: z-coordinates for the datapoints
        """
        self.add_locs(xs,ys,ts,zs)
        self.add_data(data)
        
    def get_num_points(self):
        """
        Gives the number of stations or datapoints in the dataset
        
        Returns:
            num: number of datapoints, None if there are not datapoints.
        """
        params=[self.xs,self.ys,self.zs,self.ts,self.data]
        for i,l in enumerate(params):
            if isinstance(l,(list,np.ndarray)):
                if i==len(params)-1:
                    return len(l)/self.get_size_per_point()
                else:
                    return len(l)
        return None
    
    def check_ref(self,ref):
        """
        Checks if a station has the same timesteps as the rest of the stations
        so itcan be taken as reference
        
        Returns:
            ref: name of reference station or datapoint
        """
        for sta in self.names:
            tref=self.ts[self.names==ref].tolist()
            newtref=self.ts[self.names==sta].tolist()
            if not newtref in tref:
                return False
        return True
            
    
    def add_ref(self,sta):
        """
        Adds station or datapoint as a reference for the rest of the datapoints
        
        Parameters:
            sta: name of the possible reference station
        """
        if self.ref_possible():
            if self.names is None:
                raise Exception('Names have not been defined')
            else:
                if not sta in self.names:
                    raise Exception('The station does not exist')
        else:
            raise Exception('This type of data cannot be referenced')
            
        if self.refs==None and self.ts==None:
            self.refs=[sta]
        elif not self.refs==None:
            if sta in self.refs:
                raise Exception('The station has already been included as a reference')
        elif self.ts==None:
            self.refs.append(sta)
        else:
            if not self.check_ref(sta):
                raise Exception('The new reference do not have the same time steps as the rest of the stations')
            else:
                if self.refs==None:
                    self.refs=[sta]
                else:
                    self.refs.append(sta)
    
    def remove_ref(self,sta):
        """
        Removes station or datapoint as a reference for the rest of the datapoints
        
        Parameters:
            sta: name of reference station
        """
        if not self.refs==None:
            if not sta in self.refs:
                raise Exception('This station has not been set as a reference') 
            elif len(self.refs)==1:
                self.refs=None
            else:
                self.refs.remove(sta)
        else:
            print('References have not been set')
    
    def ref_possible(self):
        """
        Defines if references can be taken for certain datatype
        
        Returns:
            possible: references can be taken for the datatype (default True but can be override)
        """
        return True
    
    def get_locs(self):
        """
        Gives the locations for the datapoints
        
        Returns:
            xs: x-coordinates
            ys: y-coordinates
            ts: times for datapoints
            zs: z-coordinates
        """
        return self.xs,self.ys,self.ts,self.zs
    
    def get_data(self,unravel=True):
        """
        Gives the referenced deformation datapoints
        
        Parameters:
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
        
        Returns:
            data: referenced deformation
        """
        return self.reference_dataset(self.data,unravel)
        
    def reference_dataset(self,data,unravel=True):
        """
        References the deformation according to the reference stations
        
        Parameters:
            data: data to be referenced
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
        
        Returns:
            data: referenced deformation
        """
        if self.refs is None:
            ref_data=data
        elif self.ts is None:
            ref_data=np.copy(data)
            for i in range(len(self.comps)):
                ref=np.mean([ref_data[i*len(self.xs):(i+1)*len(self.xs)][self.names==r] for r in self.refs])
                ref_data[i*len(self.xs):(i+1)*len(self.xs)]-=ref
        else:
            tref=self.ts[self.names==self.refs[0]]
            ref_data=np.copy(data)
            for i in range(len(self.comps)):
                for t in tref:
                    ref=0
                    for r in self.refs:
                        pos=np.argwhere(np.logical_and(self.ts==t,self.names==r))[0]
                        ref+=data[i*len(self.xs)+pos]/len(self.refs)
                    for sta in self.names:
                        stapos=np.argwhere(np.logical_and(self.ts==t,self.names==sta))[0]
                        ref_data[i*len(self.xs)+stapos]-=ref
        
        if not unravel:
            ref_comps=()
            for i in range(len(self.comps)):
                ref_comps+=(ref_data[i*len(self.xs):(i+1)*len(self.xs)],)
            ref_data=ref_comps

        return ref_data
    
    def reference_errors(self,err,unravel=True):
        """
        References the uncertainties using error propagation
        
        Parameters:
            err: uncertainties to be referenced
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
        
        Returns:
            err: referenced uncertainties
        """
        if self.refs is None:
            ref_err=err
        elif self.ts is None:
            ref_err=np.copy(err)
            for i in range(len(self.comps)):
                ref=np.sum([ref_err[i*len(self.xs):(i+1)*len(self.xs)][self.names==r]**2 for r in self.refs])
                ref_err[i*len(self.xs):(i+1)*len(self.xs)]=np.sqrt((ref_err[i*len(self.xs):(i+1)*len(self.xs)]**2+ref)/(len(self.refs)+1))
        else:
            tref=self.ts[self.names==self.refs[0]]
            ref_err=np.copy(err)
            for i in range(len(self.comps)):
                for t in tref:
                    ref=0
                    for r in self.refs:
                        pos=np.argwhere(np.logical_and(self.ts==t,self.names==r))[0]
                        ref+=err[i*len(self.xs)+pos]**2
                    for sta in self.names:
                        stapos=np.argwhere(np.logical_and(self.ts==t,self.names==sta))[0]
                        ref_err[i*len(self.xs)+stapos]=np.sqrt((ref_err[i*len(self.xs)+stapos]**2+ref)/(len(self.refs)+1))
        
        if not unravel:
            ref_comps=()
            for i in range(len(self.comps)):
                ref_comps+=(ref_err[i*len(self.xs):(i+1)*len(self.xs)],)
            ref_err=ref_comps

        return ref_err
    
    def from_model(self,func,offsets=None,unravel=True):
        """
        Uses the function from the forward model and add the offsets if they are computed
        
        Parameters:
            func: forward model function
            offsets: None or array that contains the offset for each component
            unravel (boolean): If True will give a single list with all the deformation in the datapoints
            if False, it will several array depending on the number of components in the dataset.
            
        Returns:
            ux: modeled deformation in the x-axis
            uy: modeled deformation in the y-axis
            uz: modeled deformation in the vertical
        """
        if 'time' in func.__name__:
            funct=lambda x,y: func(x,y,self.ts)
            model=self.from_model3d(funct)
        else:
            model=self.from_model3d(func)
        new_model=np.array([])
        if not offsets is None:
            if not len(model)==len(offsets):
                raise Exception('Not enough number of offsets')
            for i,comp in enumerate(model):
                comp+=offsets[i]
                
        for comp in model:
            new_model=np.concatenate((new_model,comp))
            
        ref_model=self.reference_dataset(new_model,unravel)
        
        return ref_model
    
    def get_errors(self):
        """
        Gives the referenced uncertainties for the datapoints
        
        Returns:
            errors: referenced uncertainties
        """
        return self.reference_errors(self.err)


from __future__ import division
import numpy as np
from . import Source
from .. import util
import time

eps = 1e-14 #numerical constant

class Okada(Source):
    """
    A class used to represent a tensile or slip dislocation with Okada (1985) model.

    Attributes
    ----------
    type: str
        describes if dislocation is tensile (open) or slip (slip)
    
    parameters : array
        names for the parameters in the model
    """
    def set_type(self,typ):
        """
        Defines the type of dislocation.
          
        Parameters:
            str: slip to represent faults or open to represent sill/dikes.
        """
        self.type=typ

    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Okada"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=5100000
        burnin=100000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Okada")
        if self.type=='slip':
            print("\txcen = %f" % x[0])
            print("\tycen = %f" % x[1])
            print("\tdepth = %f" % x[2])
            print("\tlength= %f" % x[3])
            print("\twidth = %f" % x[4])
            print("\tslip = %f" % x[5])
            print("\tstrike= %f" % x[6])
            print("\tdip = %f" % x[7])
            print("\trake = %f" % x[8])
        elif self.type=='open':
            print("\txcen = %f" % x[0])
            print("\tycen = %f" % x[1])
            print("\tdepth = %f" % x[2])
            print("\tlength= %f" % x[3])
            print("\twidth = %f" % x[4])
            print("\topening = %f" % x[5])
            print("\tstrike= %f" % x[6])
            print("\tdip = %f" % x[7])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        if self.type=='slip':
            self.parameters=("xcen","ycen","depth","length","width","slip","strike","dip","rake")
        elif self.type=='open':
            self.parameters=("xcen","ycen","depth","length","width","opening","strike","dip")
    
    def get_args(self,args, tilt):
        """
        Function that arranges the parameters for the dislocation model depending on the type.
        
        Parameters:
            args (list) : parameters given by the user
            tilt (boolean) : compute tilt displacements (True) or 3d displacements (False)
        
        Returns:
            rargs (list) : parameters for the Okada model.
        """
        nu=0.25
        if self.type=='slip':
            xcen,ycen,depth,length,width,slip,strike,dip,rake=args
            opening=0.0
        else:
            xcen,ycen,depth,length,width,opening,strike,dip=args
            slip=0.0
            rake=0.0
        rargs=[xcen, ycen,depth, length, width,slip, opening,strike, dip, rake,nu, tilt]
        return rargs
        
    # =====================
    # Forward Models
    # =====================
    def model(self,x,y,*args):
        """
        3d displacement field on surface for dislocation (Okada, 1985)

        Parameters:
            args (list) : parameters given by the user
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        rargs=self.get_args(args,tilt=False)
        return self.model_gen(x,y, *rargs)
    
    def model_tilt(self,x,y,*args):
        """
        Tilt displacement field on surface for dislocation (Okada, 1985)

        Parameters:
            args (list) : parameters given by the user
        
        Returns:
            dx (array) : tilt displacements in the x-axis (radians).
            dy (array) : tilt displacements in the y-axis (radians).
        """
        rargs=self.get_args(args,tilt=True)
        return self.model_gen(x,y, *rargs)
    
    def model_gen(self,x,y, xcen=0, ycen=0,
                        depth=5e3, length=1e3, width=1e3,
                        slip=0.0, opening=10.0,
                        strike=0.0, dip=0.0, rake=0.0,
                        nu=0.25,tilt=False):
        """
        Computes tilt or 3d displacement field on surface for dislocation (Okada, 1985)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of dislocation center (m)
            ycen: y-offset of dislocation center (m)
            depth: depth to dislocation center (m)
            length: length of dislocation path (m)
            width: width of dislocation path (m)
            slip: fault movement (m)
            opening: amount of closing or opening of sill/dike (m)
            strike: horizontal clock wise orientation from north of dislocation (degrees)
            dip: dipping angle of dislocation (degrees)
            rake: fault's angle of rupture where 0 represents a strike-slip fault and 90 represents a normal fault (degrees)
            nu: Poisson's ratio
            tilt: boolean to indicate calculation of tilt displacements (True) or 3d displacements (False) (default False)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        e = x - xcen
        n = y - ycen

        # A few basic parameter checks
        if not (0.0 <= strike <= 360.0) or not (0 <= dip <= 90):
            print('Strike',strike)
            print('Dip',dip)
            print('Please use 0<strike<360 clockwise from North')
            print('And 0<dip<90 East of strike convention')
            raise ValueError

        # Don't allow faults that prech the surface
        d_crit = width/2 * np.sin(np.deg2rad(dip))
        nans=np.array([e*np.nan,e*np.nan,e*np.nan])
        if depth<d_crit:
            return nans
        elif length<0:
            return nans
        elif width<0:
            return nans
        elif rake>180:
            return nans
        elif not -1.0 <= nu <= 0.5:
            return nans
        
        #assert depth >= d_crit, 'depth must be greater than {}'.format(d_crit)
        #assert length >=0, 'fault length must be positive'
        #assert width >=0, 'fault length must be positive'
        #assert rake <= 180, 'rake should be:  rake <= 180'
        #assert -1.0 <= nu <= 0.5, 'Poisson ratio should be: -1 <= nu <= 0.5'

        strike = np.deg2rad(strike) #transformations accounted for below
        dip    = np.deg2rad(dip)
        rake   = np.deg2rad(rake)

        L = length
        W = width

        U1 = np.cos(rake) * slip
        U2 = np.sin(rake) * slip
        U3 = opening

        d = depth + np.sin(dip) * W / 2 #fault top edge
        ec = e + np.cos(strike) * np.cos(dip) * W / 2
        nc = n - np.sin(strike) * np.cos(dip) * W / 2
        x = np.cos(strike) * nc + np.sin(strike) * ec + L / 2
        y = np.sin(strike) * nc - np.cos(strike) * ec + np.cos(dip) * W
        p = y * np.cos(dip) + d * np.sin(dip)
        q = y * np.sin(dip) - d * np.cos(dip)

        if tilt:
            ssx=Okada.dx_ss
            dsx=Okada.dx_ds
            tfx=Okada.dx_tf
            
            ssy=Okada.dy_ss
            dsy=Okada.dy_ds
            tfy=Okada.dy_tf
        else:
            ssx=Okada.ux_ss
            dsx=Okada.ux_ds
            tfx=Okada.ux_tf
            
            ssy=Okada.uy_ss
            dsy=Okada.uy_ds
            tfy=Okada.uy_tf
            
            uz = - U1 / (2 * np.pi) * Okada.chinnery(Okada.uz_ss, x, p, L, W, q, dip, nu) - \
                   U2 / (2 * np.pi) * Okada.chinnery(Okada.uz_ds, x, p, L, W, q, dip, nu) + \
                   U3 / (2 * np.pi) * Okada.chinnery(Okada.uz_tf, x, p, L, W, q, dip, nu)
            
        ux = - U1 / (2 * np.pi) * Okada.chinnery(ssx, x, p, L, W, q, dip, nu) - \
               U2 / (2 * np.pi) * Okada.chinnery(dsx, x, p, L, W, q, dip, nu) + \
               U3 / (2 * np.pi) * Okada.chinnery(tfx, x, p, L, W, q, dip, nu)

        uy = - U1 / (2 * np.pi) * Okada.chinnery(ssy, x, p, L, W, q, dip, nu) - \
               U2 / (2 * np.pi) * Okada.chinnery(dsy, x, p, L, W, q, dip, nu) + \
               U3 / (2 * np.pi) * Okada.chinnery(tfy, x, p, L, W, q, dip, nu)


        ue = np.sin(strike) * ux - np.cos(strike) * uy
        un = np.cos(strike) * ux + np.sin(strike) * uy
        
        if tilt:
            return ue,un
        else:
            return ue,un,uz

    def chinnery(f, x, p, L, W, q, dip, nu):
        """
        Chinnery's notation [equation (24) p. 1143]
        """
        u =  (f(x, p, q, dip, nu) -
              f(x, p - W, q, dip, nu) -
              f(x - L, p, q, dip, nu) +
              f(x - L, p - W, q, dip, nu))

        return u


    def ux_ss(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        u = xi * q / (R * (R + eta)) + \
            Okada.I1(xi, eta, q, dip, nu, R) * np.sin(dip)
        k = (q != 0)
        #u[k] = u[k] + np.arctan2( xi[k] * (eta[k]) , (q[k] * (R[k])))
        u[k] = u[k] + np.arctan( (xi[k] * eta[k]) / (q[k] * R[k]) )
        return u


    def uy_ss(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        u = (eta * np.cos(dip) + q * np.sin(dip)) * q / (R * (R + eta)) + \
            q * np.cos(dip) / (R + eta) + \
            Okada.I2(eta, q, dip, nu, R) * np.sin(dip)
        return u


    def uz_ss(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        u = db * q / (R * (R + eta)) + \
            q * np.sin(dip) / (R + eta) + \
            Okada.I4(db, eta, q, dip, nu, R) * np.sin(dip)
        return u
    
    def dx_ss(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        u = -xi * q**2 * Okada.A(eta,R) * np.cos(dip) + \
            (xi * q / R**3 - Okada.K1(xi, eta, q, dip, nu, R)) * np.sin(dip)
        return u
    
    def dy_ss(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        yb = eta * np.cos(dip) + q * np.sin(dip)
        
        u = db * q / R**3 * np.cos(dip) + \
            (xi**2 * q * Okada.A(eta,R) * np.cos(dip) - np.sin(dip) / R + yb * q / R**3 - Okada.K2(xi, eta, q, dip, nu, R)) * np.sin(dip)
        return u


    def ux_ds(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        u = q / R - \
            Okada.I3(eta, q, dip, nu, R) * np.sin(dip) * np.cos(dip)
        return u


    def uy_ds(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        u = ( (eta * np.cos(dip) + q * np.sin(dip)) * q / (R * (R + xi)) -
               Okada.I1(xi, eta, q, dip, nu, R) * np.sin(dip) * np.cos(dip) )
        k = (q != 0)
        u[k] = u[k] + np.cos(dip) * np.arctan( (xi[k] * eta[k]) / (q[k] * R[k]))
        return u


    def uz_ds(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        u = ( db * q / (R * (R + xi)) -
              Okada.I5(xi, eta, q, dip, nu, R, db) * np.sin(dip) * np.cos(dip) )
        k = (q != 0)
        #u[k] = u[k] + np.sin(dip) * np.arctan2(xi[k] * eta[k] , q[k] * R[k])
        u[k] = u[k] + np.sin(dip) * np.arctan( (xi[k] * eta[k]) / (q[k] * R[k]))
        return u
    
    def dx_ds(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        
        u = db * q / R**3 + \
            q * np.sin(dip) / (R * (R + eta)) + \
            Okada.K3(xi, eta, q, dip, nu, R) * np.sin(dip) * np.cos(dip)
        
        return u
    
    def dy_ds(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi ** 2 + eta ** 2 + q ** 2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        yb = eta * np.cos(dip) + q * np.sin(dip)
        
        u = yb * db * q * Okada.A(xi,R) - \
            (2 * db / (R * (R + xi)) + xi * np.sin(dip) / (R * (R + eta))) * np.sin(dip) + \
            Okada.K1(xi, eta, q, dip, nu, R) * np.sin(dip) * np.cos(dip)
        
        return u


    def ux_tf(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi**2 + eta**2 + q**2)
        u = q**2 / (R * (R + eta)) - \
            (Okada.I3(eta, q, dip, nu, R) * np.sin(dip)**2)
        return u


    def uy_tf(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi**2 + eta**2 + q**2)
        u = - (eta * np.sin(dip) - q * np.cos(dip)) * q / (R * (R + xi)) - \
            (np.sin(dip) * xi * q / (R * (R + eta))) - \
            (Okada.I1(xi, eta, q, dip, nu, R) * np.sin(dip) ** 2)
        k = (q != 0)
        #u[k] = u[k] + np.sin(dip) * np.arctan2(xi[k] * eta[k] , q[k] * R[k])
        u[k] = u[k] + np.sin(dip) * np.arctan( (xi[k] * eta[k]) / (q[k] * R[k]) )
        return u


    def uz_tf(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi**2 + eta**2 + q**2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        u = (eta * np.cos(dip) + q * np.sin(dip)) * q / (R * (R + xi)) + \
             np.cos(dip) * xi * q / (R * (R + eta)) - \
             Okada.I5(xi, eta, q, dip, nu, R, db) * np.sin(dip)**2
        k = (q != 0)
        u[k] = u[k] - np.cos(dip) * np.arctan( (xi[k] * eta[k]) / (q[k] * R[k]) )
        return u
    
    def dx_tf(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi**2 + eta**2 + q**2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        yb = eta * np.cos(dip) + q * np.sin(dip)
        
        u = q**2 * np.sin(dip) / R**3 - q**3 * Okada.A(eta,R) * np.cos(dip) + Okada.K3(xi, eta, q, dip, nu, R) * (np.sin(dip))**2
        
        return u
    
    def dy_tf(xi, eta, q, dip, nu):
        """
        Auxiliary function for Okada (1985) model.
        """
        R = np.sqrt(xi**2 + eta**2 + q**2)
        db = eta * np.sin(dip) - q * np.cos(dip)
        yb = eta * np.cos(dip) + q * np.sin(dip)
        
        u = (yb * np.sin(dip) + db * np.cos(dip)) * q**2 * Okada.A(xi,R) + \
            xi * q**2 * Okada.A(eta,R) * np.sin(dip) * np.cos(dip) - \
            (2 * q/(R * (R + xi)) - Okada.K1(xi, eta, q, dip, nu, R)) * (np.sin(dip))**2
        
        return u


    def I1(xi, eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        db = eta * np.sin(dip) - q * np.cos(dip)
        if np.cos(dip) > eps:
            I = (1 - 2 * nu) * (- xi / (np.cos(dip) * (R + db))) - \
                np.sin(dip) / np.cos(dip) * Okada.I5(xi, eta, q, dip, nu, R, db)
        else:
            I = -(1 - 2 * nu)/2 * xi * q / (R + db)**2
        return I


    def I2(eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        I = (1 - 2 * nu) * (-np.log(R + eta)) - \
            Okada.I3(eta, q, dip, nu, R)
        return I


    def I3(eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        yb = eta * np.cos(dip) + q * np.sin(dip)
        db = eta * np.sin(dip) - q * np.cos(dip)
        if np.cos(dip) > eps:
            I = (1 - 2 * nu) * (yb / (np.cos(dip) * (R + db)) - np.log(R + eta)) + \
                np.sin(dip) / np.cos(dip) * Okada.I4(db, eta, q, dip, nu, R)
        else:
            I = (1 - 2 * nu) / 2 * (eta / (R + db) + yb * q / (R + db) ** 2 - np.log(R + eta))
        return I


    def I4(db, eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        if np.cos(dip) > eps:
            I = (1 - 2 * nu) * 1.0 / np.cos(dip) * \
                (np.log(R + db) - np.sin(dip) * np.log(R + eta))
        else:
            I = - (1 - 2 * nu) * q / (R + db)
        return I


    def I5(xi, eta, q, dip, nu, R, db):
        """
        Auxiliary function for Okada (1985) model.
        """
        X = np.sqrt(xi**2 + q**2)
        if np.cos(dip) > eps:
            I = (1 - 2 * nu) * 2 / np.cos(dip) * \
                 np.arctan( (eta * (X + q*np.cos(dip)) + X*(R + X) * np.sin(dip)) /
                            (xi*(R + X) * np.cos(dip)) )
            I[xi == 0] = 0
        else:
            I = -(1 - 2 * nu) * xi * np.sin(dip) / (R + db)
        return I
    
    def K1(xi, eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        db = eta * np.sin(dip) - q * np.cos(dip)
        if np.cos(dip) > eps:
            K = (1 - 2 * nu) * xi / np.cos(dip) * (1/(R * (R + db)) - np.sin(dip) / (R * (R + eta)))
        else:
            K = (1 - 2 * nu) * xi * q / (R * (R + db)**2)
        return K
    
    def K2(xi, eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        K3 = Okada.K3(xi, eta, q, dip, nu, R)
        
        K = (1 - 2 * nu) * (-np.sin(dip) / R + q * np.cos(dip) / (R * (R + eta))) - K3
        
        return K
    
    def K3(xi, eta, q, dip, nu, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        yb = eta * np.cos(dip) + q * np.sin(dip)
        db = eta * np.sin(dip) - q * np.cos(dip)
        if np.cos(dip) > eps:
            K = (1 - 2 * nu) / np.cos(dip) * ((q / R) * (1 / (R + eta)) - yb / (R * (R + db)))
        else:
            K = (1 - 2 * nu) * np.sin(dip) / (R + dp) * (xi**2 / (R * (R + dp)) - 1)
        return K
    
    def A(xieta, R):
        """
        Auxiliary function for Okada (1985) model.
        """
        A = (2 * R + xieta) / (R**3 * (R + xieta)**2)
        
        return A

    




import numpy as np
from .. import util
import scipy
from . import Source

class Nish(Source):
    """
    A class used to represent an open conduit with Nishimura (2009) model.

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Nishimura"

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Nishimura")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\tradius = %f" % x[3])
        print("\theight = %f" % x[4])
        print("\tdP= %f" % x[5])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        if self.data.ts is None:
            #return "xcen","ycen","depth","radius",'lenght','pressure','tau'
            self.parameters=("xcen","ycen","depth","radius",'lenght','pressure')
        else:
            self.parameters=("xcen","ycen","depth","radius",'lenght','pressure','dtime')

    # =====================
    # Forward Models
    # =====================
    
    def model(self, x, y, xcen, ycen, d, a, h, dP, mu=4e9, rho=2500, nu=0.25):
        """
        3d displacement field on surface from open conduit (Nishimura, 2009)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of open conduit (m)
            ycen: y-offset of open conduit (m)
            d: depth of magma column (m)
            a: conduit radius (m)
            h: height of magma column (m)
            dP: change in pressure (Pa)
            mu: shear modulus for medium (Pa) (default 4e9)
            rho: host rock density (kg/m^3) (default 2500)
            nu: poisson ratio (default 0.25)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        x=x-xcen
        y=y-ycen
        if isinstance(d,float) or isinstance(d,int):
            if h/2>d:
                c1=0
            else:
                c1=d-h/2
        else:
            c1=d-h/2
            c1[c1<0]=0
        c2=d+h/2
        #m=dP*a**2/mu
        m=1
        #m=1+h
        r=np.sqrt(x**2+y**2)
        R2=np.sqrt(r**2+c2**2)
        R1=np.sqrt(r**2+c1**2)
        angle=np.arctan2(y,x)
        g=9.8
        #s=(dP/(c2-c1)-rho*g)*a**2/(4*mu)
        s=0

        urn=(m*r/(c2-c1))*(0.5*(c2**2/R2**3-c1**2/R1**3)+(c1/(2*r**2))*(c2**3/R2**3-c1**3/R1**3)-nu*(1/R2-1/R1)-((1+nu)*c1/r**2)*(c2/R2-c1/R1))
        urs=s*r*(1/R1-1/R2-(2*nu-1)*(1/(R2+c2)-1/(R1+c1)))
        ur= urn+urs
        uzn=(-1)*m/(2*(c2-c1))*((-1)*(c2**2)*(c2-c1)/R2**3+(2*nu-1)*(c2/R2-c1/R1)-2*nu*c1*(1/R2-1/R1)-(2*nu-1)*np.log((c2+R2)/(c1+R1)))
        uzs=-s*(-c2/R2+c1/R1+(2*nu-1)*np.log((R2+c2)/(R1+c1)))
        uz=uzn+uzs
        
        ux=ur*np.cos(angle)
        uy=ur*np.sin(angle)
        
        return ux,uy,uz
    
    def model_tilt(self, x, y, xcen, ycen, d, a, h, dP, mu=4e9, rho=2500, nu=0.25):
        """
        Tilt displacement field on surface from open conduit (Nishimura, 2009)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of open conduit (m)
            ycen: y-offset of open conduit (m)
            d: depth of magma column (m)
            a: conduit radius (m)
            h: height of magma column (m)
            dP: change in pressure (Pa)
            mu: shear modulus for medium (Pa) (default 4e9)
            rho: host rock density (kg/m^3) (default 2500)
            nu: poisson ratio (default 0.25)
        
        Returns:
            dx (array) : inclination in x-axis in radians.
            dy (array) : inclination in y-axis in radians.
        """
        x=x-xcen
        y=y-ycen
        if isinstance(d,float) or isinstance(d,int):
            if h/2>d:
                c1=0
            else:
                c1=d-h/2
        else:
            c1=d-h/2
            c1[c1<0]=0
        c2=d+h/2
        m=dP*a**2/mu
        #m=0
        r=np.sqrt(x**2+y**2)
        R2=np.sqrt(r**2+c2**2)
        R1=np.sqrt(r**2+c1**2)
        angle=np.arctan2(y,x)
        g=9.8
        s=(dP/(c2-c1)-rho*g)*a**2/(4*mu)
        #s=1
        #s=tau*a/(2*mu)
        
        tiltrn=(m*r/(2*(c2-c1)))*((3*c2**2/R2**5-2*nu/R2**3)*(c2-c1)+(c2/R2**3-c1/R1**3)-(2*nu-1)*(1/(R2*(R2+c2))-1/(R1*(R1+c1))))
        tiltrs=(s/r)*(2*(1-nu)*(c2/R2-c1/R1)-(c2**3/R2**3-c1**3/R1**3))
        tiltr=tiltrn+tiltrs
        
        tiltx=tiltr*np.cos(angle)
        tilty=tiltr*np.sin(angle)
        
        return tiltx,tilty
    
    def timenbg(self,z1,ap,tp,z2):
        """
        Calculates time that takes for a column to reach certain height with no bubble growth (Nishimura, 2009)

        Parameters:
            z1 (float): final top depth of magma column (m)
            ap (float): normalized pressure
            tp (float): characteristic time (s)
            z2 (float): bottom depth of magma column (m)
        
        Returns:
            t (float) : time to reach the final top depth
        """
        return tp*(z1-1-ap*np.log(np.abs((ap-(z2-z1))/(ap-(z2-1)))))

    def depthnbg(self,t,ap,tp,z2):
        """
        Inverse function of timenbg

        Parameters:
            t (float): input time (s)
            ap (float): normalized pressure
            tp (float): characteristic time (s)
            z2 (float): bottom depth of magma column (m)
        
        Returns:
            z (float) : top depth at given time (m)
        """
        diff=lambda z1,tx: (tx-self.timenbg(z1,ap,tp,z2))**2
        if isinstance(t,float):
            res = scipy.optimize.minimize(diff, 1.0, args=(t), method='Nelder-Mead', tol=1e-6)
            z=res.x[0]
        else:
            z=np.zeros(t.shape)
            for i,t_value in enumerate(t):
                res = scipy.optimize.minimize(diff, 1.0, args=(t_value), method='Nelder-Mead', tol=1e-6)
                z[i]=res.x[0]
        return z
    
    
    def depthbg(self,t,c10,td,z2):
        """
        Function that gives top depth for magma column for a magma with bubble growth.

        Parameters:
            t (float): input time (s)
            c10 (float): initial top depth for magma column (m)
            td (float): characteristic time (s)
            z2 (float): bottom depth of magma column (m)
        
        Returns:
            z (float) : top depth at given time (m)
        """
        h0=z2-c10
        h=h0*(1+(t/td)**(3/2))
        return z2-h
    
    def depthbr(self,t,c10,tr,z2):
        """
        Function that gives top depth for magma column for a magma with bubble rising.

        Parameters:
            t (float): input time (s)
            c10 (float): initial top depth for magma column (m)
            td (float): characteristic time (s)
            z2 (float): bottom depth of magma column (m)
        
        Returns:
            z (float) : top depth at given time (m)
        """
        h0=z2-c10
        h=h0*(1-(t/tr)*5/3)**(-3/5)
        return z2-h
    
    def model_t(self, x, y, t, xcen, ycen, d, a, h, dP, td, nu=0.25, mu=4e9,rho=2500,model='nb'):
        """
        3d displacement field on surface from open conduit with time dependency(Nishimura, 2009)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of open conduit (m)
            ycen: y-offset of open conduit (m)
            d: depth of magma column (m)
            a: conduit radius (m)
            h: height of magma column (m)
            dP: change in pressure (Pa)
            td: characteristic time (s)
            nu: poisson ratio (default 0.25)
            mu: shear modulus for medium (Pa) (default 4e9)
            rho: host rock density (kg/m^3) (default 2500)
            model: magma rheology model, no bubble growth (nb), bubble growth (bg) and bubble rising (br) (default nb)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        g=9.8
        if model=='nb':
            ap=dP/(rho*g*(d-h/2))
            z1=(d-h/2)*self.depthnbg(t,ap,td,(d+h/2)/(d-h/2))
        elif model=='bg':
            z1=self.depthbg(t,d-h/2,td,d+h/2)
        elif model=='br':
            z1=self.depthbr(t,d-h/2,td,d+h/2)
            
        z2=d+(h/2)
        h1=z2-z1
        d1=z2-h1/2
        ux,uy,uz=self.model(x, y, xcen, ycen, d1, a, h1, dP)
        ux0,uy0,uz0=self.model(x, y, xcen, ycen, d, a, h, dP)
        return ux-ux0,uy-uy0,uz-uz0



from __future__ import division
import numpy as np
from .. import util
import scipy
from . import Source

class Yang(Source):
    """
    Class that represents a pressurized well following the implementation from Yang et., al. (1988).

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Yang"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=5100000
        burnin=100000
        thin=1000
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Yang")
        print("\txcen = %f" % x[0])
        print("\tycen = %f" % x[1])
        print("\tz0 = %f" % x[2])
        print("\tP= %f" % x[3])
        print("\ta = %f" % x[4])
        print("\tb = %f" % x[5])
        print("\tphi = %f" % x[6])
        print("\ttheta= %f" % x[7])
        print("\tmu = %f" % x[8])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","pressure","a","b","az","dip")
        
    ##residual function for least_squares
    def fun(self, x):
        ux, uy, uz = self.forward(xcen=x[0],ycen=x[1],z0=x[2],P=x[3],a=x[4],b=x[5],phi=x[6],theta=x[7],mu=x[8])
        diff =np.concatenate((ux,uy,uz))-self.get_obs()
        return diff

    # =====================
    # Forward Models
    # =====================
    
    def model(self,x,y,xcen=0,ycen=0,z0=5e3,P=10,a=2,b=1,phi=0,theta=0,mu=9.6e9,nu=0.25):
        """
        Computes surface deformation due to a prolate spheroid pressurized chamber in elastic half-space
        Yang et al., vol 93, JGR, 4249-4257, 1988)     

        Parameters:
            x: x-coordinate (m)
            y: y-coordinate (m)
            xcen: source easting epicenter (m)
            ycen: source northing epicenter (m)
            z0: source depth (m)
            P: excess pressure (Pa)
            a: major axis [km]
            b: minor axis [km]
            phi: strike (degrees clockwise from north)
            theta: dip (degrees from horizontal)
            mu: normalized shear modulus (Pa)
            nu: poisson ratio (unitless)
        
        Returns:
            ux: deformation in the x-axis (m)
            uy: deformation in the y-axis (m)
            uz: deformation in the z-axis (m)
        """
        
        d_crita = a * np.sin(np.deg2rad(theta))
        d_critb = b * np.cos(np.deg2rad(theta))
        
        nans=np.array([x*np.nan,x*np.nan,x*np.nan])
        if d_crita>=z0 or b>=z0:
            return nans
        if theta==0:
            theta=0.1
        elif theta==90:
            theta=89.9

        tp = 0 #topography vector?
        phi = np.deg2rad(phi) 
        theta = np.deg2rad(theta)  

        # Store some commonly used parameters (material properties)
        coeffs = np.zeros(3)
        coeffs[0] = 1 / (16 * mu * (1 - nu))
        coeffs[1] = 3 - 4 * nu
        coeffs[2] = 4 * (1 - nu) * (1 - 2 * nu)    
        # Elastic constant array
        matrl = np.array([mu,mu,nu])   

        # Geometery
        e_theta = np.zeros(2)
        e_theta[0] = np.sin(theta)
        e_theta[1] = np.cos(theta)
        cosp = np.cos(phi)
        sinp = np.sin(phi)
        c = np.sqrt(a ** 2 - b ** 2)

        xn = x - xcen
        yn = y - ycen

        # Call spheroid function to get geometric paramters
        # NOTE: had to add file name
        sph = Yang.spheroid(a,b,c,matrl,phi,theta,P)

        # Rotate points
        xp = xn * cosp - yn * sinp
        yp = yn * cosp + xn * sinp

        # Run forward model ?why called twice?, for each side of model...
        xi = c
        Up1,Up2,Up3 = Yang.yang(sph,xi,z0,xp,yp,0,matrl,e_theta,coeffs,tp)
        xi = -xi
        Um1,Um2,Um3 = Yang.yang(sph,xi,z0,xp,yp,0,matrl,e_theta,coeffs,tp)

        # Sum
        U1r = -Up1 + Um1
        U2r = -Up2 + Um2

        # Rotate horiz. displacements back to the orig. coordinate system
        U1 = U1r * cosp + U2r * sinp
        U2 = -U1r * sinp + U2r * cosp
        U3 = Up3 - Um3

        return U1,U2,U3
    
    def model_depth(self,x,y,z,xcen=0,ycen=0,z0=5e3,P=10,a=2,b=1,phi=0,theta=0,mu=9.6e9,nu=0.25):
        """
        Computes deformation at depth due to a prolate spheroid pressurized chamber in elastic half-space
        Yang et al., vol 93, JGR, 4249-4257, 1988)     

        Parameters:
            x: x-coordinate (m)
            y: y-coordinate (m)
            z: depth-coordinate (m)
            xcen: source easting epicenter (m)
            ycen: source northing epicenter (m)
            z0: source depth (m)
            P: excess pressure (Pa)
            a: major axis [km]
            b: minor axis [km]
            phi: strike (degrees clockwise from north)
            theta: dip (degrees from horizontal)
            mu: normalized shear modulus (Pa)
            nu: poisson ratio (unitless)
        
        Returns:
            ux: deformation in the x-axis (m)
            uy: deformation in the y-axis (m)
            uz: deformation in the z-axis (m)
        """
        if theta==0:
            theta=0.1
        elif theta==90:
            theta=89.9

        tp = 0 #topography vector?
        phi = np.deg2rad(phi) 
        theta = np.deg2rad(theta)  

        # Store some commonly used parameters (material properties)
        coeffs = np.zeros(3)
        coeffs[0] = 1 / (16 * mu * (1 - nu))
        coeffs[1] = 3 - 4 * nu
        coeffs[2] = 4 * (1 - nu) * (1 - 2 * nu)    
        # Elastic constant array
        matrl = np.array([mu,mu,nu])   

        # Geometery
        e_theta = np.zeros(2)
        e_theta[0] = np.sin(theta)
        e_theta[1] = np.cos(theta)
        cosp = np.cos(phi)
        sinp = np.sin(phi)
        c = np.sqrt(a ** 2 - b ** 2)

        xn = x - xcen
        yn = y - ycen

        # Call spheroid function to get geometric paramters
        # NOTE: had to add file name
        sph = Yang.spheroid(a,b,c,matrl,phi,theta,P)

        # Rotate points
        xp = xn * cosp - yn * sinp
        yp = yn * cosp + xn * sinp

        # Run forward model ?why called twice?, for each side of model...
        xi = c
        Up1,Up2,Up3 = Yang.yang(sph,xi,z0,xp,yp,z,matrl,e_theta,coeffs,tp)
        xi = -xi
        Um1,Um2,Um3 = Yang.yang(sph,xi,z0,xp,yp,z,matrl,e_theta,coeffs,tp)

        # Sum
        U1r = -Up1 + Um1
        U2r = -Up2 + Um2

        # Rotate horiz. displacements back to the orig. coordinate system
        U1 = U1r * cosp + U2r * sinp
        U2 = -U1r * sinp + U2r * cosp
        U3 = Up3 - Um3
        

        return U1,U2,U3

    def spheroid(a,b,c,matrl,phi,theta,P):
        ''' 
        Geometry used in yang pressure source computation 
        '''
        pi = np.pi    
        lamda = matrl[0]
        mu = matrl[1]
        nu = matrl[2]

        ac = (a - c) / (a + c)
        L1 = np.log(ac)
        iia = 2 / a / c ** 2 + L1 / c ** 3
        iiaa = 2 / 3 / a ** 3 / c ** 2 + 2 / a / c ** 4 + L1 / c ** 5
        coef1 = -2 * pi * a * b ** 2
        Ia = coef1 * iia
        Iaa = coef1 * iiaa

        u = 8 * pi * (1 - nu)
        Q = 3 / u
        R = (1 - 2 * nu) / u

        a11 = 2 * R * (Ia - 4 * pi)
        a12 = -2 * R * (Ia + 4 * pi)
        a21 = Q * a ** 2 * Iaa + R * Ia - 1
        a22 = -(Q * a ** 2 * Iaa + Ia * (2 * R - Q))

        coef2 = 3 * lamda + 2 * mu
        w = 1 / (a11 * a22 - a12 * a21)
        e11 = (3 * a22 - a12) * P * w / coef2
        e22 = (a11 - 3 * a21) * P * w / coef2

        Pdila = 2 * mu * (e11 - e22)
        Pstar = lamda * e11 + 2 * (lamda + mu) * e22
        a1 = -2 * b ** 2 * Pdila
        b1 = 3 * b ** 2 * Pdila / c ** 2 + 2 * (1 - 2 * nu) * Pstar

        sph = np.zeros(10)
        sph[0] = a
        sph[1] = b
        sph[2] = c
        sph[3] = phi
        sph[4] = theta
        sph[5] = Pstar
        sph[6] = Pdila
        sph[7] = a1
        sph[8] = b1
        sph[9] = P
        return sph


    def yang(sph,xi,z0,x,y,z,matrl,e_theta,coeffs,tp):
        """
        Auxiliary function for spheroidal model
        """
        pi = np.pi
        
        # Load required spheroid parameters
        a = sph[0]
        b = sph[1]
        c = sph[2]
        #phi=sph[3]
        #theta=sph[4]
        #Pstar=sph[5]
        Pdila = sph[6]
        a1 = sph[7]
        b1 = sph[8]
        #P=sph[9]

        sinth = e_theta[0]
        costh = e_theta[1]

        # Poisson's ratio, Young's modulus, and the Lame coeffiecents mu and lamda
        nu = matrl[2]
        nu4 = coeffs[1]
        #nu2=1 - 2 * nu
        nu1 = 1 - nu
        coeff = a * b ** 2 / c ** 3 * coeffs[0]

        # Introduce new coordinates and parameters (Yang et al., 1988, page 4251):
        xi2 = xi * costh
        xi3 = xi * sinth
        y0 = 0
        z00 = tp + z0
        x1 = x
        x2 = y - y0
        x3 = z - z00
        xbar3 = z + z00
        y1 = x1
        y2 = x2 - xi2
        y3 = x3 - xi3
        ybar3 = xbar3 + xi3
        r2 = x2 * sinth - x3 * costh
        q2 = x2 * sinth + xbar3 * costh
        r3 = x2 * costh + x3 * sinth
        q3 = -x2 * costh + xbar3 * sinth
        rbar3 = r3 - xi
        qbar3 = q3 + xi
        R1 = (y1 ** 2 + y2 ** 2 + y3 ** 2) ** (0.5)
        R2 = (y1 ** 2 + y2 ** 2 + ybar3 ** 2) ** (0.5)

        #C0 = y0 * costh + z00 * sinth # check this?
        C0=z0/sinth

        betatop = (costh * q2 + (1 + sinth) * (R2 + qbar3))
        betabottom = costh * y1
        # Strange replacement for matlab 'find'
        #nz=np.flatnonzero(np.abs(betabottom) != 0) #1D index
        nz  = (np.abs(betabottom) != 0) # 2D index
        atnbeta = pi / 2 * np.sign(betatop)
        # change atan --> arctan, and -1 not needed for 2D boolean index
        #atnbeta[(nz-1)]=np.atan(betatop[(nz-1)] / betabottom[(nz-1)])
        atnbeta[nz] = np.arctan(betatop[nz] / betabottom[nz])

        # Set up other parameters for dipping spheroid (Yang et al., 1988, page 4252):
        # precalculate some repeatedly used natural logs:
        Rr = R1 + rbar3
        Rq = R2 + qbar3
        Ry = R2 + ybar3
        lRr = np.log(Rr)
        lRq = np.log(Rq)
        lRy = np.log(Ry)

        # Note: dot products should in fact be element-wise multiplication
        #A1star=a1 / (R1.dot(Rr)) + b1 * (lRr + (r3 + xi) / Rr)
        #Abar1star=- a1 / (R2.dot(Rq)) - b1 * (lRq + (q3 - xi) / Rq)
        A1star =  a1 / (R1*Rr) + b1*(lRr + (r3 + xi)/Rr)
        Abar1star = -a1 / (R2*Rq) - b1*(lRq + (q3 - xi)/Rq)
        A1 = xi / R1 + lRr
        Abar1 = xi / R2 - lRq
        #A2=R1 - r3.dot(lRr)
        #Abar2=R2 - q3.dot(lRq)
        A2 = R1 - r3*lRr
        Abar2 = R2 - q3*lRq
        A3 = xi * rbar3 / R1 + R1
        Abar3 = xi * qbar3 / R2 - R2

        #B=xi * (xi + C0) / R2 - Abar2 - C0.dot(lRq)
        B = xi * (xi + C0)/R2 - Abar2 - C0*lRq
        Bstar = a1 / R1 + 2 * b1 * A2 + coeffs[1] * (a1 / R2 + 2 * b1 * Abar2)
        F1 = 0
        F1star = 0
        F2 = 0
        F2star = 0


        if z != 0:
            F1 = (-2*sinth*z* (xi*(xi+C0)/R2**3 +
                              (R2+xi+C0)/(R2*(Rq)) +
                              4*(1-nu)*(R2+xi)/(R2*(Rq))
                              )
                 )

            F1star = (2*z*(costh*q2*(a1*(2*Rq)/(R2**3*(Rq)**2) - b1*(R2 + 2*xi)/(R2*(Rq)**2)) +
                           sinth*(a1/R2**3 -2*b1*(R2 + xi)/(R2* (Rq)))
                           )
                     )

            F2 = -2*sinth*z*(xi*(xi+C0)*qbar3/R2**3 + C0/R2 + (5-4*nu)*Abar1)

            F2star = 2*z*(a1*ybar3/R2**3 - 2*b1*(sinth*Abar1 + costh*q2*(R2+xi)/(R2*Rq)))


        # Calculate little f's
        ff1 = (xi*y1/Ry +
               3/(costh)**2*(y1*lRy*sinth -y1*lRq + 2*q2*atnbeta) +
               2*y1*lRq -
               4*xbar3*atnbeta/costh
              )

        ff2 = (xi*y2/Ry +
               3/(costh)**2*(q2*lRq*sinth - q2*lRy + 2*y1*atnbeta*sinth + costh*(R2-ybar3)) -
               2*costh*Abar2 +
               2/costh*(xbar3*lRy - q3*lRq)
              )

        ff3 = ((q2*lRq - q2*lRy*sinth + 2*y1*atnbeta)/costh +
                2*sinth*Abar2 + q3*lRy - xi
              )


        # Assemble into x, y, z displacements (1,2,3):
        u1 = coeff*(A1star + nu4*Abar1star + F1star)*y1

        #u2 = coeff*(sinth*(A1star*r2+(nu4*Abar1star+F1star)*q2) +
        #            costh*(Bstar-F2star) + 2*sinth*costh*z*Abar1star)
        u2 = coeff*(sinth*(A1star*r2+(nu4*Abar1star+F1star)*q2) +
                    costh*(Bstar-F2star))
        #u3 = coeff*(-costh*(Abar1star*r2+(nu4*Abar1star-F1star)*q2) +
        #            sinth*(Bstar+F2star) + 2*(costh)**2*z*Abar1star)
        u3 = coeff*(-costh*(A1star*r2+(nu4*Abar1star-F1star)*q2) +
                    sinth*(Bstar+F2star))


        u1 = u1 + 2*coeff*Pdila*((A1 + nu4*Abar1 + F1)*y1 - coeffs[2]*ff1)

        u2 = u2 + 2*coeff*Pdila*(sinth*(A1*r2+(nu4*Abar1+F1)*q2) -
                                 coeffs[2]*ff2 + 4*nu1*costh*(A2+Abar2) +
                                 costh*(A3 - nu4*Abar3 - F2))

        u3 = u3 + 2*coeff*Pdila*(costh*(-A1*r2 + (nu4*Abar1 + F1)*q2) + coeffs[2]*ff3 +
                                 4*nu1*sinth*(A2+Abar2) +
                                 sinth*(A3 + nu4*Abar3 + F2 - 2*nu4*B))
        
        return u1,u2,u3


import numpy as np
from .. import util
from . import Source

class Vsphere(Source):
    """
    Class that represents a pressurized sphere within a viscoelastic medium Bonafede and Ferrari (2009) model.

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Vsphere"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=1100000
        burnin=10000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Vsphere")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\tradius = %f" % x[3])
        print("\tdP= %f" % x[4])
    
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","radius","dP","tau")
        
    def model(self, x, y, xcen, ycen, d, rad, dP):
        """
        Initial 3d displacement field on surface for pressurized sphere in a viscoelastic medium (Bonafede and Ferrari, 2009)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dP: change in pressure (Pa)
            
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        return model_t(x,y,0, xcen, ycen, d, rad, dP)
        
    def model_t(self, x, y, t, xcen, ycen, d, rad, dP, tau, nu=0.25, mu=4e9):
        """
        3d displacement field on surface for pressurized sphere in a viscoelastic medium (Bonafede and Ferrari, 2009)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            t: input time (s)
            xcen: x-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dP: change in pressure (Pa)
            
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        a=rad
        K=5*mu/3
        
        x = x - xcen
        y = y - ycen
        
        if rad>d:
            return x*np.Inf,x*np.Inf,x*np.Inf
        
        th, r = util.cart2pol(x,y)
        A=1+2*(mu/K)
        B=2*mu**2/(K*(3*K+mu))
        alpha=(3*K+mu)/(3*K)
        taual=alpha*tau
        R=np.sqrt(r**2+d**2)
        ur1=[(r/R**3)*(1+t/tau)+(r/R**3)*(A-B*np.exp(-t/taual)+t/tau)]
        ur2=(dP*a**3)/(4*mu)
        ur=np.array([ur2*urt for urt in ur1])

        uz=[-d/R**3*(1+t/tau)-(d/R**3)*(A-B*np.exp(-t/taual)+t/tau)]
        uz=np.array([ur2*utz for utz in uz])
        
        ux, uy = util.pol2cart(th, ur)
        
        return ux.ravel(),uy.ravel(),uz.ravel()


import numpy as np
from .. import util
from . import Source

class Mogi(Source):
    """
    A class used to represent a point source using the Mogi (1958) model.

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Mogi"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        
        Returns:
            steps (int): Number of steps used in the bayesian inversions.
            burnin (int): discarded number of steps at the begining of the inversion.
            thin (int): number of steps per sample.
        """
        steps=1100000
        burnin=100000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Mogi")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\tdV= %f" % x[3])
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","dV")

    # =====================
    # Forward Models
    # =====================
    
    def model(self,x,y, xcen, ycen, d, dV, nu=0.25):
        """
        3d displacement field on surface from point source (Mogi, 1958)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: x-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium (default 0.25)
            mu: shear modulus for medium (Pa) (default 4e9)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        # Center coordinate grid on point source
        x = x - xcen
        y = y - ycen

        # Convert to surface cylindrical coordinates
        th, rho = util.cart2pol(x,y) # surface angle and radial distance
        R = np.sqrt(d**2+rho**2)     # radial distance from source

        # Mogi displacement calculation
        C = ((1-nu) / np.pi) * dV
        ur = C * rho / R**3    # horizontal displacement, m
        uz = C * d / R**3      # vertical displacement, m

        ux, uy = util.pol2cart(th, ur)
        
        return ux, uy, uz #returns tuple
        #return np.array([ux,uy,uz])
    
    def model_tilt(self, x, y, xcen, ycen, d, dV, nu=0.25):
        """
        Tilt displacement field from point source (Mogi, 1958)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: y-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium
            mu: shear modulus for medium (Pa)
            order: highest order term to include (up to 2)
            output: 'cart' (cartesian), 'cyl' (cylindrical)

        Returns:
            dx (array) : inclination in the x-axis in radians.
            dy (array) : inclination in the y-axis in radians.
        """
        
        # Center coordinate grid on point source
        x = x - xcen
        y = y - ycen

        # Convert to surface cylindrical coordinates
        th, rho = util.cart2pol(x,y) # surface angle and radial distance
        R = np.sqrt(d**2+rho**2)     # radial distance from source

        # Mogi displacement calculation
        C = ((1-nu) / np.pi) * dV
        
        dx=3*C*d*x/R**5
        dy=3*C*d*y/R**5
        
        #print(dx)
        
        return dx, dy
    
    def calc_genmax(self,t,xcen=0,ycen=0,d=4e3,dP=100e6,a=700,nu=0.25,G=30e9,
                    mu1=0.5,eta=2e16,**kwargs):
        """ Solution for spherical source in a generalized maxwell viscoelastic
        halfspace based on Del Negro et al 2009.

        Required arguments:
        ------------------
        x: x-coordinate grid (m)
        y: y-coordinate grid (m)
        t: time (s)

        Keyword arguments:
        -----------------
        xcen: y-offset of point source epicenter (m)
        ycen: y-offset of point source epicenter (m)
        d: depth to point (m)
        dV: change in volume (m^3)
        K: bulk modulus (constant b/c incompressible)
        E: Young's moduls
        G: total shear modulus (Gpa)
        mu0: fractional shear modulus (spring part)
        mu1: fractional shear modulus (dashpot part)
        eta: viscosity (Pa s)
        output: 'cart' (cartesian), 'cyl' (cylindrical)

        """
        #WARNING: mu0 != 0
        # center coordinate grid on point source
        x = self.get_xs() - xcen
        y = self.get_ys() - ycen

        # convert to surface cylindrical coordinates
        #th, r = cart2pol(x,y)
        r = np.hypot(x,y) #surface radial distance
        R = np.hypot(d,r) #radial distance from source center

        # Calculate displacements
        #E = 2.0 * G * (1+nu)
        #K = E / (3.0* (1 - 2*nu)) #bulk modulus = (2/3)*E if poisson solid
        K = (2.0*G*(1+nu)) / (3*(1-(2*nu)))
        mu0 = 1.0 - mu1
        alpha = (3.0*K) + G #recurring terms
        beta = (3.0*K) + (G*mu0)

        # Maxwell times
        try:
            tau0 = eta / (G*mu1)
        except:
            tau0 = np.inf
        tau1 = (alpha / beta) * tau0
        tau2 = tau0 / mu0

        #print('relaxation times:\nT0={}\nT1={}\nT2={}'.format(tau0,tau1,tau2))

        term1 = ((3.0*K + 4*G*mu0) / (mu0*beta))
        term2 = ((3.0 * G**2 * np.exp(-t/tau1))*(1-mu0)) / (beta*alpha)
        term3 = ((1.0/mu0) - 1) * np.exp(-t/tau2)

        A = (1.0/(2*G)) * (term1 - term2 - term3)
        C = (dP * a**3) / R**3
        ur = C * A * r
        uz = C * A * d

        return ur, uz


import numpy as np
from .. import util
import scipy
from scipy import special
from . import Source

class Penny(Source):
    """
    Class used to represent a penny-shaped crack using the Fialko (2001) model.

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Penny"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        """
        steps=110000
        burnin=10000
        thin=100
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Penny-shaped Crack:")
        print("\tx  = %f (m)" % x[0])
        print("\ty  = %f (m)" % x[1])
        print("\td  = %f (m)" % x[2])
        print("\tP_G= %f (m)" % x[3])
        print("\ta  = %f (m)" % x[4])
        
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","pressure","radius")
        
    # =====================
    # Forward Models
    # =====================
    
    def model(self,x,y,xcen,ycen,d,P_G,a,nu=0.25):
        """
        Calculates surface deformation based on pressurized penny-shaped crack
        References: Fialko

        Parameters:
            x: x-coordinate grid (m)
            y: y-coordinate grid (m)
            xcen: x-offset of penny-shaped crack (m)
            ycen: y-offset of penny-shaped crack (m)
            d: depth to penny-shaped crack (m)
            P_G: normalized pressure with shear modulus
            a: radius for penny shaped crack (m)
            nu: poisson's ratio for medium

        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """

        #
        rd=np.copy(a)

        # Center coordinate grid on point source, normalized by radius
        x = (x - xcen) / rd
        y = (y - ycen) / rd
        z = (0 - d)    / rd

        eps=1e-8

        h  = d / rd
        r  = np.sqrt(x ** 2 + y ** 2)

        csi1,w1 = self.gauleg(eps,10,41)
        csi2,w2 = self.gauleg(10,60,41)
        csi     = np.concatenate((csi1,csi2))
        wcsi    = np.concatenate((w1,w2))

        if csi.shape[0] == 1:
            csi=csi.T

        phi1,psi1,t,wt=self.psi_phi(h)

        phi=np.matmul(np.sin(np.outer(csi,t)) , phi1*wt)
        psi=np.matmul(np.divide(np.sin(np.outer(csi,t)), np.outer(csi,t)) - np.cos(np.outer(csi,t)),psi1*wt)
        a=csi * h
        A=np.exp((-1)*a)*(a*psi + (1 + a)*phi)
        B=np.exp((-1)*a)*((1-a)*psi - a*phi)
        Uz=np.zeros(r.shape)
        Ur=np.zeros(r.shape)

        for i in range(r.size):
            J0=special.jv(0,r[i] * csi)
            Uzi=J0*(((1-2*nu)*B - csi*(z+h)*A)*np.sinh(csi*(z+h)) +(2*(1-nu)*A - csi*(z+h)*B)*np.cosh(csi*(z+h)))
            Uz[i]=np.dot(wcsi , Uzi)
            J1=special.jv(1,r[i] * csi)
            Uri=J1*(((1-2*nu)*A + csi*(z+h)*B)*np.sinh(csi*(z+h)) + (2*(1-nu)*B + csi*(z+h)*A)*np.cosh(csi*(z+h)))
            Ur[i]=np.dot(wcsi , Uri)

        ux = rd * P_G * Ur*x / r
        uy = rd * P_G * Ur*y / r
        uz = -rd * P_G * Uz

        return np.array([ux,uy,uz])
    
    def model_depth(self,x,y,z,xcen,ycen,d,P,a,mu=1e10,nu=0.25):
        """
        Calculates deformation at depth based on pressurized penny-shaped crack
        References: Fialko

        Parameters:
            x: x-coordinate grid (m)
            y: y-coordinate grid (m)
            z: z-coordinate grid (m)
            xcen: x-offset of penny-shaped crack (m)
            ycen: y-offset of penny-shaped crack (m)
            d: depth to penny-shaped crack (m)
            P: pressure (Pa)
            a: radius penny-shaped crack (Pa)
            mu: shear modulus (Pa)
            nu: poisson's ratio for medium

        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        P_G=P/mu

        #
        rd=np.copy(a)

        # Center coordinate grid on point source, normalized by radius
        x = (x - xcen) / rd
        y = (y - ycen) / rd
        z = (0 - (d-z)) / rd

        eps=1e-8

        h  = d / rd
        r  = np.sqrt(x ** 2 + y ** 2)

        csi1,w1 = self.gauleg(eps,10,41)
        csi2,w2 = self.gauleg(10,60,41)
        csi     = np.concatenate((csi1,csi2))
        wcsi    = np.concatenate((w1,w2))

        if csi.shape[0] == 1:
            csi=csi.T

        phi1,psi1,t,wt=self.psi_phi(h)

        phi=np.matmul(np.sin(np.outer(csi,t)) , phi1*wt)
        psi=np.matmul(np.divide(np.sin(np.outer(csi,t)), np.outer(csi,t)) - np.cos(np.outer(csi,t)),psi1*wt)
        a=csi * h
        A=np.exp((-1)*a)*(a*psi + (1 + a)*phi)
        B=np.exp((-1)*a)*((1-a)*psi - a*phi)
        Uz=np.zeros(r.shape)
        Ur=np.zeros(r.shape)

        for i in range(r.size):
            J0=special.jv(0,r[i] * csi)
            Uzi=J0*(((1-2*nu)*B - csi*(z+h)*A)*np.sinh(csi*(z+h)) +(2*(1-nu)*A - csi*(z+h)*B)*np.cosh(csi*(z+h)))
            Uz[i]=np.dot(wcsi , Uzi)
            J1=special.jv(1,r[i] * csi)
            Uri=J1*(((1-2*nu)*A + csi*(z+h)*B)*np.sinh(csi*(z+h)) + (2*(1-nu)*B + csi*(z+h)*A)*np.cosh(csi*(z+h)))
            Ur[i]=np.dot(wcsi , Uri)

        ux = rd * P_G * Ur*x / r
        uy = rd * P_G * Ur*y / r
        uz = -rd * P_G * Uz

        return np.array([ux,uy,uz])

    def psi_phi(self,h):
        """
        Auxiliary function for the Fialko (2001) model
        """
        t,w=self.gauleg(0,1,41)
        t=np.array(t)
        g=-2.0*t/np.pi
        d=np.concatenate((g,np.zeros(g.size)))
        T1,T2,T3,T4=self.giveT(h,t,t)
        T1p=np.zeros(T1.shape)
        T2p=np.zeros(T1.shape)
        T3p=np.zeros(T1.shape)
        T4p=np.zeros(T1.shape)
        N=t.size
        for j in range(N):
            T1p[:,j]=w[j]*T1[:,j]
            T2p[:,j]=w[j]*T2[:,j]
            T3p[:,j]=w[j]*T3[:,j]
            T4p[:,j]=w[j]*T4[:,j]
        M1=np.concatenate((T1p,T3p),axis=1)
        M2=np.concatenate((T4p,T2p),axis=1)
        Kp=np.concatenate((M1,M2),axis=0)
        y=np.matmul(np.linalg.inv(np.eye(2*N,2*N)-(2/np.pi)*Kp),d)
        phi=y[0:N]
        psi=y[N:2*N]
        return phi,psi,t,w

    def giveP(self,h,x):
        """
        Auxiliary function for the Fialko (2001) model
        """
        P=np.zeros((4,x.size))
        P[0]=(12*np.power(h,2)-np.power(x,2))/np.power((4*np.power(h,2)+np.power(x,2)),3)
        P[1] = np.log(4*np.power(h,2)+np.power(x,2)) + (8*np.power(h,4)+2*np.power(x,2)*np.power(h,2)-np.power(x,4))/np.power(4*np.power(h,2)+np.power(x,2),2)
        P[2] = 2*(8*np.power(h,4)-2*np.power(x,2)*np.power(h,2)+np.power(x,4))/np.power(4*np.power(h,2)+np.power(x,2),3)
        P[3] = (4*np.power(h,2)-np.power(x,2))/np.power((4*np.power(h,2)+np.power(x,2)),2)
        return P

    def giveT(self,h,t,r):
        """
        Auxiliary function for the Fialko (2001) model
        """
        M = t.size
        N = r.size
        T1 = np.zeros((M,N)) 
        T2 = np.zeros((M,N)) 
        T3 = np.zeros((M,N))
        for i in range(M):
            Pm=self.giveP(h,t[i]-r)
            Pp=self.giveP(h,t[i]+r)
            T1[i] = 4*np.power(h,3)*(Pm[0,:]-Pp[0,:])
            T2[i] = (h/(t[i]*r))*(Pm[1,:]-Pp[1,:]) +h*(Pm[2,:]+Pp[2,:])
            T3[i] = (np.power(h,2)/r)*(Pm[3,:]-Pp[3,:]-2*r*((t[i]-r)*Pm[0,:]+(t[i]+r)*Pp[0,:]))
        T4=np.copy(T3.T)
        return T1,T2,T3,T4
    
    def gauleg(self,a,b,n):
        """
        Auxiliary function for the Fialko (2001) model
        """
        xs, cs = self.gauleg_params1(n)
        coeffp = 0.5*(b+a) 
        coeffm = 0.5*(b-a)
        ts = coeffp - coeffm*xs
        ws=cs*coeffm
        #contribs = cs*f(ts)
        #return coeffm*np.sum(contribs)
        return ts[::-1],ws
    
    def gauleg_params1(self,n):
        """
        Auxiliary function for the Fialko (2001) model
        """
        xs,cs=np.polynomial.legendre.leggauss(n)
        return xs,cs


import copy
import numpy as np
import sys
import pickle
from multiprocessing import Pool
from scipy.optimize import least_squares
from scipy.optimize import basinhopping
from scipy.optimize import differential_evolution
from scipy.optimize import shgo
import random
import string
import time
import os

os.environ["OMP_NUM_THREADS"] = "1"

class Inverse:
    """
    Class to implement inverse methods operating on one or more analytical volcanic source models
    
    Attributes:
        sources (array): sources that will be part of a model
        obs (Data): data object containing one or more datasets
        minresidual (float): minimum residual achieved with any inversion algorithm
        minparms (array): parameters corresponding to the minimum residual
    """
    def __init__(self, obs):
        self.sources = []       #simple list that will contain instances of class Source
        self.obs     = obs      #instance of Data
        #self.model   = None
        #self.iter=0
        self.steps=None
        self.burnin=None
        self.thin=None
        self.minresidual=None
        self.minparms=None
    
    #add a new source to the geometry
    def register_source(self, source):
        """
        Adds a source to the model
        
        Parameters:
            source (Source): source object to be added to the model
        """
        self.sources.append(source)

    #interface to scipy bounded nonlinear least squares implementation
    def nlsq(self):
        """
        Non-linear least squares approach using the trust reflective algorithm
        
        Returns:
            params (array): parameters for the model with the minimum residual
        """
        self.minresidual=1e6
        if len(self.sources)==0:
            raise Exception('You need to include at least one source')
        
        params = copy.deepcopy(least_squares(self.residual, self.get_x0(), bounds=self.get_bounds()))
        print(self.minresidual)
        return params
    
    def bh(self):
        """
        Optimization using basin-hopping algorithm from scipy
        
        Returns:
            params (array): parameters for the model with the minimum residual
        """
        self.minresidual=1e6
        minimizer_kwargs = dict(method="L-BFGS-B", bounds=self.get_bounds_de())
        if len(self.sources)==0:
            raise Exception('You need to include at least one source')
        
        params = copy.deepcopy(basinhopping(self.residual_bh, self.get_x0(),niter=100,minimizer_kwargs=minimizer_kwargs))
        print(self.minresidual)
        return params
    
    def de(self):
        """
        Optimization using differential evolution algorithm from scipy
        
        Returns:
            params (array): parameters for the model with the minimum residual
        """
        self.minresidual=1e6
        if len(self.sources)==0:
            raise Exception('You need to include at least one source')
        
        params = copy.deepcopy(differential_evolution(self.residual_bh, bounds=self.get_bounds_de()))
        print(self.minresidual)
        return params
    
    def shg(self):
        """
        Optimization using simplicial homology global algorithm from scipy
        
        Returns:
            params (array): parameters for the model with the minimum residual
        """
        self.minresidual=1e6
        if len(self.sources)==0:
            raise Exception('You need to include at least one source')
        start=time.time()
        params = copy.deepcopy(shgo(self.residual_bh, bounds=self.get_bounds_de()))
        end=time.time()
        print('Time:',end-start)
        print(self.minresidual)
        return params
    
    def log_prior(self,theta):
        """
        Logarithmic uniform prior probabilities for the model parameters
        
        Parameters:
            theta (array): parameter values
            
        Returns:
            log_prior (float): logarithm of the prior probability 
        """
        j=0
        if not -10<theta[-1]<10:
            return -np.inf
        
        for k,source in enumerate(self.sources):
            parnames=source.get_parnames()
            for i in range(source.get_num_params()):
                low,high,ini=self.par2log(source,i)
                if not low<theta[j]<high:
                    return -np.inf
                j+=1
        
        return 0.0

    def log_likelihood(self, theta):
        """
        Calculate the log likelihood with a normal distribution.

        Parameters:
            theta (array): parameter values

        Returns:
            likeli (float): logarithm of the likelihood 
        """
        model = self.get_model(theta[0:-1])
        data = self.obs.get_data()
        diff = data - model
        std_devs = self.obs.get_errors()
        log_std_devs = np.log(std_devs)

        likeli = -0.5 * (len(model) * np.log(2 * np.pi) + np.sum(2 * log_std_devs) + np.sum(((diff/std_devs) ** 2)))
        
        if np.isnan(likeli):
            return -np.inf
        
        return likeli
    
    def log_probability(self,theta):
        """
        Logarithm of the probability for a set of parameters
        
        Parameters:
            theta (array): parameter values
            
        Returns:
            log_prob (float): logarithm of the probability for the set of parameters 
        """
        lp = self.log_prior(theta)
        if not np.isfinite(lp):
            return -np.inf
        likeli=self.log_likelihood(theta)
        #print('lp',lp)
        #print('likeli',likeli)
        #print('likeli',np.isnan(likeli))
        return lp + likeli
    
    def mcmc_em(self,name=None,move=None):
        """
        Bayesian inversion approach with multiple algorithms using the emcee library
        
        Parameters:
            name (str): filename for the h5 file that will contain the traces
            move (str): update algorithm for the steps (metropolis, stretch, kde, de, desnooker, redblue)
            
        Returns:
            traces (array): traces that will give the posterior distribution for the parameters
        """
        import emcee
        
        inis=[]
        for k,source in enumerate(self.sources):
            parnames=source.get_parnames()
            for i in range(source.get_num_params()):
                low,high,ini=self.par2log(source,i)
                inis.append(ini)
        inis.append(0.0)
        
        print(inis)
        
        steps,burnin,thin=self.get_numsteps()
        
        pos = np.array(inis) + 1e-4 * np.random.randn(2*len(inis), len(inis))
        nwalkers, ndim = pos.shape

        if move=='metropolis':
            moves=emcee.moves.GaussianMove(1.0)
        elif move=='stretch':
            moves=emcee.moves.StretchMove()
        elif move=='kde':
            moves=emcee.moves.KDEMove()
        elif move=='de':
            moves=emcee.moves.DEMove()
        elif move=='desnoooker':
            moves=emcee.moves.DESnookerMove()
        elif move=='redblue':
            moves=emcee.moves.RedBlueMove()
        
        backend = emcee.backends.HDFBackend(name+'.h5')
        backend.reset(nwalkers, ndim)
        
        with Pool() as pool:
            if move is None:
                sampler = emcee.EnsembleSampler(nwalkers, ndim, self.log_probability, pool=pool, backend=backend)
            else:
                sampler = emcee.EnsembleSampler(nwalkers, ndim, self.log_probability, pool=pool,moves=moves, backend=backend)
            sampler.run_mcmc(pos, int(steps/nwalkers),skip_initial_state_check=True, progress=True)
        
        traces = sampler.get_chain(discard=int(burnin/nwalkers), thin=int(thin/nwalkers), flat=True)
        
        traces=traces.T.tolist()[0:-1]
        
        traces,labels=self.traces2lin(traces)
        
        if name is None:
            name=''.join(random.choices(string.ascii_lowercase, k=5))
        
        solution=dict()
        for i,trace in enumerate(traces):
            solution[labels[i]]=trace
        
        with open(name+'.pkl', 'wb') as f:
            pickle.dump(solution, f)
        
        return traces
    
    def mcmc(self,name=None):
        """
        Bayesian inversion approach with multiple algorithms using the pymc library
        
        Parameters:
            name (str): filename for the h5 file that will contain the traces
            
        Returns:
            traces (array): traces that will give the posterior distribution for the parameters
            MDL (pymc.MCMC): inversion object
        """
        import pymc
        self.minresidual=1e6
        data=self.obs.get_data()
        errors=self.obs.get_errors()
        
        #############THIS PART IS FOR REGULARIZATION#################
        reg=False
        for source in self.sources:
            if source.reg:
                reg=True
                break
        
        if reg:
            data=data.tolist()
            errors=errors.tolist()
            for source in self.sources:
                if source.reg:
                    data+=[0.0 for i in range(source.get_num_params())]
                    errors+=[1.0 for i in range(source.get_num_params())]
            data=np.array(data)
            errors=np.array(errors)
        #############THIS PART IS FOR REGULARIZATION#################
        
        if self.obs.err is None:
            wts=1.0
        else:
            wts=1.0/errors
        def model(data):
            #Distribution for every parameter
            theta=[]
            orders=[]
            for k,source in enumerate(self.sources):
                parnames=source.get_parnames()
                for i in range(source.get_num_params()):
                    low,high,ini=self.par2log(source,i)
                    thetat=pymc.Uniform(parnames[i]+str(k),low,high,value=ini)
                    theta.append(thetat)
            model_pars=theta
            #sigma=pymc.Uniform('sigma',0,100,value=0.5)
            sigma=pymc.Uniform('sigma',-10,10,value=0)
            #sigma=-7

            #Deformation is the deterministic variable
            @pymc.deterministic(plot=False)
            def defo(model_pars=model_pars):
                return self.get_model(model_pars)

            #Probability distribution
            z = pymc.Normal('z', mu=defo*wts, tau=1.0/10**sigma, value=data*wts, observed=True)
            return locals()
        
        #Making MCMC model
        MDL = pymc.MCMC(model(data))

        #Choosing Adaptive Metropolis-Hastings for step method
        MDL.use_step_method(pymc.AdaptiveMetropolis,MDL.model_pars)

        #Steps, burnin and thining for each model
        steps,burnin,thin=self.get_numsteps()

        #Number of runs
        MDL.sample(steps,burnin,thin) 

        #Getting the traces
        traces=[]
        for k,source in enumerate(self.sources):
            parnames=source.get_parnames()
            for i in range(source.get_num_params()):
                traces.append(MDL.trace(parnames[i]+str(k)))
        traces=np.array(traces)
        traces,labels=self.traces2lin(traces)
        
        if name is None:
            name=''.join(random.choices(string.ascii_lowercase, k=5))
        
        solution=dict()
        for i,trace in enumerate(traces):
            solution[labels[i]]=trace
        #solution['MDL']=MDL
        
        with open(name+'.pkl', 'wb') as f:
            pickle.dump(solution, f)
        
        return traces,MDL
    
    def doublelog(self,x):
        """
        Double logarithmic function to sweep the parameter space more efficiently
        
        Parameters:
            x (float): parameter value 
            
        Returns:
            logx (float): logarithm in base 10 of the parameter value 
            if value > 1e-2 or value < -1e-2 if not it will give a linear function
        """
        if x>=-1e-2 and x<=1e-2:
            m=2.0/0.02
            return m*x
        elif x<-1e-2:
            return -np.log10(-x)-3
        else:
            return np.log10(x)+3
    
    def invdoublelog(self,y):
        """
        Inverse for double logarithmic function to get the parameter value
        
        Parameters:
            logx (float): value given by the doublelog function
            
        Returns:
            x (float): parameter value
        """
        if y>=-1 and y<=1:
            m=0.02/2.0
            return m*y
        elif y<-1:
            return -10**(-(y+3.0))
        else:
            return 10**(y-3.0)
    
    def par2log(self,source,i):
        """
        Normalizes the bounds and initial guess for the parameters  
        
        Parameters:
            source (Source): source where the parameters will be normalized
            i (int): index for the parameter
            
        Returns:
            low (float): normalized lower bound for the parameter
            high (float): normalized upper bound for the parameter
            ini (float): normalized initial guess for the parameter
        """
        #orders,parnames=self.get_parnames_orders()
        parnames=source.get_parnames()
        order=int(np.log10(np.max([np.abs(source.low_bounds[i]),np.abs(source.high_bounds[i])])))-1
        if not parnames[i]=='pressure':
            low=source.low_bounds[i]/(10**order)
            high=source.high_bounds[i]/(10**order)
            ini=source.x0[i]/(10**order)
        else:
            low=self.doublelog(source.low_bounds[i])
            high=self.doublelog(source.high_bounds[i])
            ini=self.doublelog(source.x0[i])
            
        return low,high,ini
    
    def par2lin(self,pars):
        """
        Multiplies normalized parameters with the correspondant magnitude  
        
        Parameters:
            pars (array): normalized parameters
            
        Returns:
            linpars (array): normalized parameters multiply by their correspondant magnitude
        """
        parnames,orders=self.get_parnames_orders()
        
        linpars=[]
        for i in range(len(pars)):
            par=pars[i]
            if parnames[i]=='pressure':
                linpars.append(self.invdoublelog(par))
            else:
                linpars.append(par*orders[i])
        return linpars
    
    def traces2lin(self,traces):
        """
        Multiplies normalized parameters with the correspondant magnitude
        for all the values in the traces
        
        Parameters:
            traces (array): traces with normalized values 
            
        Returns:
            data (array): normalized traces multiply by their correspondant magnitude
            labels (array): names for the parameters
        """
        data=[]
        labels=[]
        parnames,orders=self.get_parnames_orders()
        for i,trace in enumerate(traces):
            temp=np.array(trace[:])
            if parnames[i]=='pressure':
                vdata=np.array([self.invdoublelog(temp[j]) for j in range(len(temp))])
                data.append(vdata)
            else:
                data.append(temp*orders[i])
            labels.append(parnames[i])
        data=np.vstack(data)
        return data,labels
    
    def get_parnames_orders(self):
        """
        Gives the magnitudes for the parameters
        
        Returns:
            parnames (array): names for the parameters
            orders (array): magnitudes for the parameters
        """
        if len(self.sources)==1:
            source=self.sources[0]
            return source.get_parnames(),source.get_orders()
        else:
            parnamest=[]
            orderst=[]
            for k,source in enumerate(self.sources):
                parnames=source.get_parnames()
                orders=source.get_orders()
                for j,name in enumerate(parnames):
                    parnamest.append(name+str(k))
                    orderst.append(orders[j])
            
            return parnamest,orderst
    
    def set_numsteps(self,steps,burnin,thin):
        """
        Set the number of steps, discarded steps and steps 
        per sample for a Bayesian inversion, if the model has 
        more than one source the default is
        steps=6600000, burnin=6000000, thin=1000, if not it will
        take the values defined in the source object
        
        Returns:
            steps (int): number of steps 
            burnin (int): number of initial discarded steps
            thin (int): number of steps per sample
        """
        self.steps=steps
        self.burnin=burnin
        self.thin=thin
    
    def get_numsteps(self):
        """
        Gives the number of steps, discarded steps and steps 
        per sample for a Bayesian inversion, if the model has 
        more than one source the default is
        steps=6600000, burnin=6000000, thin=1000, if not it will
        take the values defined in the source object
        
        Returns:
            steps (int): number of steps 
            burnin (int): number of initial discarded steps
            thin (int): number of steps per sample
        """
        if not (self.steps is None or self.burnin is None or self.thin is None):
            steps=self.steps
            burnin=self.burnin
            thin=self.thin
        elif len(self.sources)>1:
            steps=6600000
            burnin=600000
            thin=1000
        else:
            source=self.sources[0]
            steps,burnin,thin=source.bayesian_steps()
        return steps,burnin,thin
    
    #initial guess of source model characteristics, defined when creating the source model
    def get_x0(self):
        """
        Concatenates the initial values for the parameters for all sources
        in the model
        
        Returns:
            x0 (array): intial values for the parameters
        """
        x0 = []
        
        for s in self.sources:
            x0 = np.concatenate((x0, s.x0))

        return x0

    #high and low bounds for the parameters
    def get_bounds(self):
        """
        Concatenates and returns the lower and upper bounds for the parameters for all sources
        in the model
        
        Returns:
            low_b (array): lower bounds for the parameters
            high_b (array): upper bounds for the parameters
        """
        low_b  = []
        high_b = []

        for s in self.sources:
            low_b  = np.concatenate((low_b,  s.low_bounds))
            high_b = np.concatenate((high_b, s.high_bounds))

        return (low_b, high_b)
    
    def get_bounds_de(self):
        """
        Arranges lower and upper bounds in one array for the differential evolution algorithm
        
        Returns:
            bounds (array): bounds for the parameters
        """
        bounds  = []

        for s in self.sources:
            for i in range(len(s.high_bounds)):
                if len(bounds)==0:
                    bounds=[(s.low_bounds[i],s.high_bounds[i])]
                else:
                    bounds = np.concatenate((bounds,[(s.low_bounds[i],s.high_bounds[i])]))

        return bounds
    
    def forward(self, x, unravel=True):
        """
        Computes the forward model for single or multisource model
        
        Parameters:
            unravel (boolean): if True it will return just one array,
            if False it will separate the deformation in components
        
        Returns:
            data (array): modeled deformation
        """
        param_cnt = 0
        data=None
        for s in self.sources:
            datat = s.forward(x[param_cnt:param_cnt+s.get_num_params()],unravel)
            param_cnt += s.get_num_params()
            
            if data is None:
                data=datat
            else:
                if unravel:
                    data += datat
                else:
                    for i,comp in enumerate(data):
                        comp+=datat[i]

        return data
    
    def residual(self,x):
        """
        Computes the residuals for the non-linear least squares algorithm
        
        Parameters:
            x (array): parameters in an iteration
            
        Returns:
            res (float): residual for the iteration
        """
        if self.obs.get_errors() is None:
            res=self.obs.get_data()-self.forward(x)
        else:
            res=(self.obs.get_data()-self.forward(x))/self.obs.get_errors()
        rest=np.sqrt(np.sum(res**2))
        if self.minresidual>rest:
            self.minresidual=rest
            self.minparms=x
        
        return res
    
    def residual_bh(self,x):
        """
        Computes the residuals for the optimization algorithms
        
        Parameters:
            x (array): parameters in an iteration
            
        Returns:
            rest (float): residual for the iteration
        """
        if self.obs.get_errors() is None:
            res=self.obs.get_data()-self.forward(x)
        else:
            res=(self.obs.get_data()-self.forward(x))/self.obs.get_errors()
        rest=np.sqrt(np.sum(res**2))
        if self.minresidual>rest:
            self.minresidual=rest
            self.minparms=x
        
        return rest
    
    def inv_opening(self,xcen,ycen,depth,length,width,strike,dip,reg=False,lamb=1):
        """
        Inversion for opening values in a discretized sill
        
        Parameters:
            xcen (float): x-coordinate location for center of the sill (m)
            ycen (float): y-coordinate location for center of the sill (m)
            depth (float): depth for center of the sill (m)
            length (float): length of the discretized sill (m)
            width (float): width of the discretized sill (m)
            strike (float): azimuth angle for the sill clockwise from north (degrees)
            dip (float): dipping angle for the sill (degrees)
            reg (boolean): if True apply regularization, if False regularization
            won't be applied
            lamb (float): regularization constant 
            
        Returns:
            ops (array): openings for the sill patches
        """
        s=self.sources[0]
        G=s.get_greens(xcen,ycen,depth,length,width,strike,dip)
        if reg==True:
            d=np.array(self.obs.get_data().tolist()+np.zeros((s.ln*s.wn,)).tolist())
            L=s.get_laplacian(xcen,ycen,depth,length,width,strike,dip)
            #print(L)
            newG=np.concatenate((G,lamb*L),axis=0)
            ops=np.linalg.lstsq(newG, d, rcond=None)[0]
        else:
            ops=np.linalg.lstsq(G, self.obs.get_obs(), rcond=None)[0]
        return ops
    
    def get_params_openings(self,xcen,ycen,depth,length,width,strike,dip,ln,wn,ops):
        """
        Gives the parameters for each sill patch
        
        Parameters:
            xcen (float): x-coordinate location for center of the sill (m)
            ycen (float): y-coordinate location for center of the sill (m)
            depth (float): depth for center of the sill (m)
            length (float): length of the discretized sill (m)
            width (float): width of the discretized sill (m)
            strike (float): azimuth angle for the sill clockwise from north (degrees)
            dip (float): dipping angle for the sill (degrees)
            ln (int): divisions in length
            wn (int): divisions in width
            ops (array): openings for sill patches 
            
        Returns:
            params (array): parameters for sill patches
        """
        s=self.sources[0]
        xcs,ycs,zcs=s.get_centers(xcen,ycen,depth,length,width,strike,dip,ln,wn)
        params=[]
        slength=length/ln
        swidth=width/wn
        for i in range(len(xcs)):
            params+=[xcs[i],ycs[i],zcs[i],slength,swidth,ops[i],strike,dip]
        return params

    def get_model(self,x):
        """
        Computes the deformation values for a given set of parameters
        
        Parameters:
            x (array): normalized values for the parameters 
            
        Returns:
            defo (array): deformation values for the given parameters
        """
        defo=None
        rmod=None
        param_cnt = 0
        xlin=[]
        linparst=self.par2lin(x)
        
        for s in self.sources:
            linpars=linparst[param_cnt:param_cnt+s.get_num_params()]
            #############THIS PART IS FOR REGULARIZATION#################
            if s.reg:
                if rmod is None:
                    rmod=(s.get_laplacian()@np.array(linpars)).tolist()
                else:
                    rmod+=(s.get_laplacian()@np.array(linpars)).tolist()
            #############THIS PART IS FOR REGULARIZATION#################
            
            if defo is None:
                defo=s.forward(linpars)
            else:
                defo+=s.forward(linpars)
            param_cnt += s.get_num_params()
        
        self.residual(linparst)
        
        #############THIS PART IS FOR REGULARIZATION#################
        if not rmod is None:
            defo=np.array(defo.tolist()+rmod)
        #############THIS PART IS FOR REGULARIZATION#################
        
        return defo
    
    ##output writers
    def print_model(self):
        """
        Prints the parameter names and their values
        """
        param_cnt = 0

        for s in self.sources:
            s.print_model(self.model.x[param_cnt:param_cnt+s.get_num_params()])
            param_cnt += s.get_num_params()

    ##writes gmt files for horizontal and vertical deformation, each, to use with gmt velo.
    def write_forward_gmt(self, prefix, volcano=None):
        if self.model is not None:

            ux,uy,uz = self.forward(self.model.x)

            dat = np.zeros(self.obs.data['id'].to_numpy().size, 
                dtype=[ ('lon', float), ('lat', float), ('east', float), ('north', float), 
                        ('esig', float), ('nsig', float), ('corr', float), ('id', 'U6')] )

            dat['lon']   = self.obs.data['lon'].to_numpy()
            dat['lat']   = self.obs.data['lat'].to_numpy()
            dat['east']  = ux*1000
            dat['north'] = uy*1000
            dat['esig']  = ux*0
            dat['nsig']  = ux*0
            dat['corr']  = ux*0
            dat['id']    = self.obs.data['id'].to_numpy()

            print(dat)

            #horizontal predictions    
            np.savetxt(prefix+"_hori.gmt", dat, fmt='%s' )

            #vertical predictions    
            dat['east']  = ux*0
            dat['north'] = uz*1000
            np.savetxt(prefix+"_vert.gmt", dat, fmt='%s' )
            
            if volcano is not None:
                import utm
                dat = np.zeros(len(self.sources), dtype=[ ('lon', float), ('lat', float), ('id', 'U6')] )

                print(len(self.sources))                    

                param_cnt = 0
                source_cnt = 0

                for s in self.sources:
                    e_loc = volcano[0] + self.model.x[param_cnt]
                    n_loc = volcano[1] + self.model.x[param_cnt+1]

                    param_cnt += s.get_num_params()

                    loc_ll = utm.to_latlon(e_loc, n_loc, volcano[2], volcano[3])

                    dat["lat"][source_cnt] = loc_ll[0]
                    dat["lon"][source_cnt] = loc_ll[1]
                    dat["id"][source_cnt]  = s.get_source_id()

                    source_cnt += 1

                print(dat)                    

                np.savetxt(prefix+"_source_loc.gmt", dat, fmt='%s' )
        else:
            print("No model, nothing to write")


import numpy as np
from .. import util
import scipy
from scipy.integrate import quad,quad_vec
from . import Source

class Mctigue(Source):
    """
    A class used to represent a spherical source using the McTigue (1987) implementation

    Attributes
    ----------
    parameters : array
        names for the parameters in the model
    """
    def get_source_id(self):
        """
        The function defining the name for the model.
          
        Returns:
            str: Name of the model.
        """
        return "Mctigue"
    
    def bayesian_steps(self):
        """
        Function that defines the number of steps for a bayesian inversion.
        """
        steps=1100000
        burnin=10000
        thin=1000
        
        return steps,burnin,thin

    def print_model(self, x):
        """
        The function prints the parameters for the model.
        
        Parameters:
           x (list): Parameters for the model.
        """
        print("Mctigue")
        print("\tx = %f" % x[0])
        print("\ty = %f" % x[1])
        print("\td = %f" % x[2])
        print("\tradius = %f" % x[3])
        print("\tdV= %f" % x[4])
    
    def set_parnames(self):
        """
        Function defining the names for the parameters in the model.
        """
        self.parameters=("xcen","ycen","depth","radius","dV")
    
    # =====================
    # Forward Models
    # =====================
    
    def model(self, x, y, xcen, ycen, d, rad, dV, nu=0.25, mu=4e9):
        """
        3d displacement field on surface from spherical source (McTigue, 1987)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: y-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium (default 0.25)
            mu: shear modulus for medium (Pa) (default 4e9)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        # center coordinate grid on point source
        x = x - xcen
        y = y - ycen
        if rad>d:
            return x*np.Inf,x*np.Inf,x*np.Inf
        
        dP = dV*mu/(np.pi*rad**3)

        # dimensionless scaling term
        scale = dP * d / mu
        eps = rad / d #NOTE eps = 0.5 # mctigue fig3

        # convert to surface cylindrical coordinates
        th, r = util.cart2pol(x,y)
        r = r / d #dimensionless radial distance

        # 1st order mctigue is essentially Mogi solution
        uz = eps**3 * ((1-nu) * (1 / np.hypot(r,1)**3))
        ur = eps**3 * ((1-nu) * (r / np.hypot(r,1)**3))

        # 2nd order term
        A = ((1 - nu) * (1 + nu)) / (2 * (7 - 5*nu))
        B = (15 * (2 - nu) * (1 - nu)) / (4 * (7 - 5*nu))
        uz2 =  -eps**6 * ((A * (1 / np.hypot(r,1)**3)) - (B * (1 / np.hypot(r,1)**5)))
        ur2 =  -eps**6 * ((A * (r / np.hypot(r,1)**3)) - (B * (r / np.hypot(r,1)**5)))
        uz += uz2
        ur += ur2

        # Convert back to dimensional variables
        uz = uz * scale
        ur = ur * scale

        # Convert surface cylindrical to cartesian
        ux, uy = util.pol2cart(th, ur)
        return ux, uy, uz
    
    def model_tilt(self, x, y, xcen, ycen, d, rad, dV, nu=0.25, mu=4e9):
        """
        Tilt displacement field from spherical source (McTigue, 1987)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            xcen: y-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium
            mu: shear modulus for medium (Pa)
            order: highest order term to include (up to 2)
            output: 'cart' (cartesian), 'cyl' (cylindrical)

        Returns:
            dx (array) : inclination in the x-axis in radians.
            dy (array) : inclination in the y-axis in radians.
        """
        # center coordinate grid on point source
        x = x - xcen
        y = y - ycen
        if rad>d:
            return x*np.Inf,x*np.Inf,x*np.Inf
        
        dP = dV*mu/(np.pi*rad**3)

        # dimensionless scaling term
        scale = dP * d / mu
        eps = rad / d #NOTE eps = 0.5 # mctigue fig3

        # convert to surface cylindrical coordinates
        th, r = util.cart2pol(x,y)
        r = r / d #dimensionless radial distance

        # 1st order mctigue is essentially Mogi solution
        dx = 3 * (x/(d**2)) * eps**3 * ((1-nu) * (1 / np.hypot(r,1)**5))
        dy = 3 * (y/(d**2)) * eps**3 * ((1-nu) * (1 / np.hypot(r,1)**5))
        #print(dx)

        # 2nd order term
        A = ((1 - nu) * (1 + nu)) / (2 * (7 - 5*nu))
        B = (15 * (2 - nu) * (1 - nu)) / (4 * (7 - 5*nu))
        
        dx2 =  -eps**6 * (x/(d**2)) * ((A * (3 / np.hypot(r,1)**5)) - (5 * B * (1 / np.hypot(r,1)**7)))
        dy2 =  -eps**6 * (y/(d**2)) * ((A * (3 / np.hypot(r,1)**5)) - (5 * B * (1 / np.hypot(r,1)**7)))
        
        #print(dx2)
        
        dx += dx2
        dy += dy2
        
        dx = dx*scale
        dy = dy*scale
        
        return dx, dy
    
    def model_depth(self, x, y, z, xcen, ycen, d, rad, dV, nu=0.25, mu=4e9):
        """
        3d displacement field at depth from dislocation point source (McTigue, 1987)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            z: z-coordinate for displacement (m)
            xcen: y-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium (default 0.25)
            mu: shear modulus for medium (Pa) (default 4e9)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        # center coordinate grid on point source
        x = x - xcen
        y = y - ycen
        if isinstance(d,float):
            if rad>d:
                return x*np.Inf,x*np.Inf,x*np.Inf
        else:
            if len(d[rad>d])>0:
                return x*np.Inf,x*np.Inf,x*np.Inf
        
        dP = dV*mu/(np.pi*rad**3)

        # dimensionless scaling term
        scale = dP * d / mu
        eps = rad / d #NOTE eps = 0.5 # mctigue fig3

        # convert to surface cylindrical coordinates
        th, r = util.cart2pol(x,y)
        rho = r / d #dimensionless radial distance
        zeta = z / d
        
        uz0=(eps**3)*0.25*(1-zeta)/(rho**2+(1-zeta)**2)**(1.5)
        ur0=(eps**3)*0.25*rho/(rho**2+(1-zeta)**2)**(1.5)
        
        Auz1=self.auz1(nu,rho,zeta)
        Aur1=self.aur1(nu,rho,zeta)
        
        R=np.sqrt(rho**2+(1-zeta)**2)
        sint=rho/R
        cost=(1-zeta)/R
        C3=[eps*(1+nu)/(12*(1-nu)),5*(eps**3)*(2-nu)/(24*(7-5*nu))]
        D3=[-(eps**3)*(1+nu)/12,(eps**5)*(2-nu)/(4*(7-nu))]
        P0=1
        P2=0.5*(3*cost**2-1)
        dP0=0
        dP2=3*cost
        ur38=-0.5*P0*D3[0]/R**2+(C3[1]*(5-4*nu)-1.5*D3[1]/R**2)*P2/R**2
        ut39=-(2*C3[0]*(1-nu)-0.5*D3[0]/R**2)*dP0-(C3[1]*(1-2*nu)+0.5*D3[1]/R**2)*dP2/R**2
        
        ut39=ut39*sint
        Auz3=ur38*cost-ut39*sint
        Aur3=ur38*sint+ut39*cost
        
        Auz6=self.auz6(nu,rho,zeta)
        Aur6=self.aur6(nu,rho,zeta)
        
        uz=uz0+(eps**3)*(Auz1+Auz3)+(eps**6)*Auz6
        ur=ur0+(eps**3)*(Aur1+Aur3)+(eps**6)*Aur6
       
        ux=ur*x/r
        uy=ur*y/r
        
        ux=ux*dP*d/mu
        uy=uy*dP*d/mu
        uz=uz*dP*d/mu
        
        return ux, uy, uz
    
    def stress(self, x, y, z, xcen, ycen, d, rad, dV, nu=0.25, mu=4e9):
        """
        Stress field at depth from dislocation point source (McTigue, 1987)

        Parameters:
            x: x-coordinate for displacement (m)
            y: y-coordinate for displacement (m)
            z: z-coordinate for displacement (m)
            xcen: y-offset of point source epicenter (m)
            ycen: y-offset of point source epicenter (m)
            d: depth to point (m)
            rad: chamber radius (m)
            dV: change in volume (m^3)
            dP: change in pressure (Pa)
            nu: poisson's ratio for medium (default 0.25)
            mu: shear modulus for medium (Pa) (default 4e9)
        
        Returns:
            ux (array) : displacements in east in meters.
            uy (array) : displacements in north in meters.
            uz (array) : displacements in vertical in meters.
        """
        hx=0.001*np.abs(np.max(x)-np.min(x))
        hy=0.001*np.abs(np.max(y)-np.min(y))
        if hx==0:
            h=hy
        elif hy==0:
            h=hx
        elif hx<hy:
            h=hx
        else:
            h=hy
            
        u,v,w=self.model_depth(x, y, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        
        upx,vpx,wpx=self.model_depth(x+h, y, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        umx,vmx,wmx=self.model_depth(x-h, y, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        dudx=0.5*(upx-umx)/h
        dvdx=0.5*(vpx-vmx)/h
        dwdx=0.5*(wpx-wmx)/h
        
        upy,vpy,wpy=self.model_depth(x, y+h, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        umy,vmy,wmy=self.model_depth(x, y-h, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        dudy=0.5*(upy-umy)/h
        dvdy=0.5*(vpy-vmy)/h
        dwdy=0.5*(wpy-wmy)/h
        
        double=False
        if isinstance(z,float):
            if z==0:
                double=True
            else:
                double==False
        elif len(z[z==0])>0:
            double=True
            
        if double:
            upz,vpz,wpz=self.model_depth(x, y, z+2*h, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
            umz,vmz,wmz=self.model_depth(x, y, z, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        else:
            upz,vpz,wpz=self.model_depth(x, y, z+h, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
            umz,vmz,wmz=self.model_depth(x, y, z-h, xcen, ycen, d, rad, dV, nu=nu, mu=mu)
        dudz=0.5*(upz-umz)/h
        dvdz=0.5*(vpz-vmz)/h
        dwdz=0.5*(wpz-wmz)/h
        
        sxx=2*(1+nu)*dudx*mu
        syy=2*(1+nu)*dvdy*mu
        szz=2*(1+nu)*dwdz*mu
        sxy=(1+nu)*(dudy+dvdx)*mu
        sxz=(1+nu)*(dudz+dwdx)*mu
        syz=(1+nu)*(dvdz+dwdy)*mu
        
        return np.c_[x,u,v,w,sxx,syy,szz,sxy,sxz,syz]
    
    def auz1(self,nu,r,zeta):
        """
        Auxiliary function to calculate displacements (for detail check McTigue, 1987)
        """
        R=np.sqrt(r**2+zeta**2)
        sigma=lambda tt: 0.5*tt*np.exp(-tt)
        a7=lambda tt: 0.5*sigma(tt)*(2*(1-nu)-tt*zeta)*np.exp(tt*zeta)*scipy.special.jv(0, tt*R)
        a17=lambda tt: 0.5*sigma(tt)*((1-2*nu)-tt*zeta)*np.exp(tt*zeta)*scipy.special.jv(0, tt*R)
        duz=lambda tt: a7(tt)+a17(tt)
        return quad_vec(duz,0,50)[0]
    
    def auz6(self,nu,r,zeta):
        """
        Auxiliary function to calculate displacements (for detail check McTigue, 1987)
        """
        R=np.sqrt(r**2+zeta**2)
        sigma=lambda tt: 1.5*(tt+tt**2)*np.exp(-tt)/(7-5*nu)
        tau=lambda tt: tt**2*np.exp(-tt)/(7-5*nu)
        a7=lambda tt: 0.5*sigma(tt)*(2*(1-nu)-tt*zeta)*np.exp(tt*zeta)*scipy.special.jv(0, tt*R)
        a17=lambda tt: 0.5*tau(tt)*((1-2*nu)-tt*zeta)*np.exp(tt*zeta)*scipy.special.jv(0, tt*R)
        duz=lambda tt: a7(tt)+a17(tt)
        return quad_vec(duz,0,50)[0]
    
    def aur1(self,nu,r,zeta):
        """
        Auxiliary function to calculate displacements (for detail check McTigue, 1987)
        """
        aur=r*np.nan
        R=r
        sigma=lambda tt: 0.5*tt*np.exp(-tt)
        a18=lambda tt: 0.5*sigma(tt)*(2*(1-nu)-tt*zeta)*np.exp(-tt*zeta)*scipy.special.jv(1, tt*R)
        a8=lambda tt: 0.5*sigma(tt)*((1-2*nu)-tt*zeta)*np.exp(-tt*zeta)*scipy.special.jv(1, tt*R)
        dur=lambda tt: a8(tt)+a18(tt)
        return quad_vec(dur,0,50)[0]
    
    def aur6(self,nu,r,zeta):
        """
        Auxiliary function to calculate displacements (for detail check McTigue, 1987)
        """
        aur=r*np.nan
        R=r
        sigma=lambda tt: (tt**2)*np.exp(-tt)/(7-5*nu)
        a18=lambda tt: 0.5*sigma(tt)*(2*(1-nu)-tt*zeta)*np.exp(-tt*zeta)*scipy.special.jv(1, tt*R)
        a8=lambda tt: 0.5*sigma(tt)*((1-2*nu)-tt*zeta)*np.exp(-tt*zeta)*scipy.special.jv(1, tt*R)
        dur=lambda tt: a8(tt)+a18(tt)
        return quad_vec(dur,0,50)[0]